@@ -2,8 +2,12 @@
 
 eval `dpkg-architecture -s`
 
+if [ -z "${REPO_ARCH_PATH}" ]; then
+	REPO_ARCH_PATH="`pwd`/../repo"
+fi
+
 #find the latest version we can build:
-XPRA_TAR_XZ=`ls ../pkgs/xpra-4.2.*.tar.xz | grep -v html5 | sort -V | tail -n 1`
+XPRA_TAR_XZ=`ls ../pkgs/xpra-4.3*.tar.xz | grep -v html5 | sort -V | tail -n 1`
 if [ -z "${XPRA_TAR_XZ}" ]; then
 	echo "no xpra source found"
 	exit 0
@@ -47,7 +51,11 @@ MATCH=`find $REPO_ARCH_PATH/ -name "${DEB_FILENAME}" | wc -l`
 if [ "$MATCH" != "0" ]; then
 	echo "package already exists"
 else
-	debuild -us -uc -b
+	if [ `arch` == "aarch64" ]; then
+		debuild --no-lintian -us -uc -b
+	else
+		debuild -us -uc -b
+	fi
 	ls -la ../xpra*deb
 	mv ../python*xpra*deb ../xpra*deb ../python*xpra*changes ../xpra*changes "$REPO_ARCH_PATH"
 fi
