@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -9,9 +9,10 @@ import os
 import re
 
 from xpra.client.client_widget_base import ClientWidgetBase
+from xpra.client.window_backing_base import fire_paint_callbacks
 from xpra.os_util import bytestostr, OSX, WIN32, is_Wayland
 from xpra.common import GRAVITY_STR
-from xpra.util import typedict, envbool, envint, WORKSPACE_UNSET, WORKSPACE_NAMES
+from xpra.util import net_utf8, typedict, envbool, envint, WORKSPACE_UNSET, WORKSPACE_NAMES
 from xpra.log import Logger
 
 log = Logger("window")
@@ -31,6 +32,7 @@ PROPERTIES_DEBUG = [x.strip() for x in os.environ.get("XPRA_WINDOW_PROPERTIES_DE
 SET_SIZE_CONSTRAINTS = envbool("XPRA_SET_SIZE_CONSTRAINTS", True)
 DEFAULT_GRAVITY = envint("XPRA_DEFAULT_GRAVITY", 0)
 OVERRIDE_GRAVITY = envint("XPRA_OVERRIDE_GRAVITY", 0)
+FORCE_FLUSH = envbool("XPRA_FORCE_FLUSH", False)
 
 
 class ClientWindowBase(ClientWidgetBase):
@@ -66,6 +68,7 @@ class ClientWindowBase(ClientWidgetBase):
         self._skip_taskbar = False
         self._iconified = False
         self._focused = False
+        self._opaque_region = ()
         self.window_gravity = OVERRIDE_GRAVITY or DEFAULT_GRAVITY
         self.border = border
         self.cursor_data = None
@@ -99,14 +102,15 @@ class ClientWindowBase(ClientWidgetBase):
         workspacelog("init_window(..) workspace from client properties %s: %s", self._client_properties, wn(workspace))
         if workspace is not None:
             #client properties override application specified workspace value on init only:
-            metadata[b"workspace"] = workspace
+            metadata["workspace"] = workspace
         self._window_workspace = WORKSPACE_UNSET        #will get set in set_metadata if present
         self._desktop_workspace = self.get_desktop_workspace()  #pylint: disable=assignment-from-none
         workspacelog("init_window(..) workspace=%s, current workspace=%s",
                      wn(self._window_workspace), wn(self._desktop_workspace))
-        if self.max_window_size and b"size-constraints" not in metadata:
+        if self.max_window_size and "size-constraints" not in metadata:
             #this ensures that we will set size-constraints and honour max_window_size:
-            metadata[b"size-constraints"] = {}
+            metadata.pop("workspace", None)
+            metadata["size-constraints"] = {}
         #initialize gravity early:
         sc = typedict(metadata.dictget("size-constraints", {}))
         self.window_gravity = OVERRIDE_GRAVITY or sc.intget("gravity", DEFAULT_GRAVITY)
@@ -147,12 +151,12 @@ class ClientWindowBase(ClientWidgetBase):
             "geometry-hints"        : dict(self.geometry_hints),
             "content-type"          : self.content_type,
             "attributes"            : attributes,
-            "gravity"               : GRAVITY_STR.get(self.window_gravity),
+            "gravity"               : GRAVITY_STR.get(self.window_gravity, self.window_gravity),
             #"border"                : self.border or "",
             #cursor_data
             "max-size"              : self.max_window_size,
             "button-state"          : self.button_state,
-            "offset"                : self.window_offset,
+            "offset"                : self.window_offset or (0, 0),
             })
         return info
 
@@ -172,6 +176,8 @@ class ClientWindowBase(ClientWidgetBase):
         self._backing.border = self.border
         self._backing.default_cursor_data = self.default_cursor_data
         self._backing.gravity = self.window_gravity
+        #this is only used by cairo to request a repaint for the fps counter:
+        self._backing.repaint = self.repaint
         return self._backing._backing
 
 
@@ -216,20 +222,20 @@ class ClientWindowBase(ClientWidgetBase):
             self.update_icon(current_icon)
 
     def update_icon(self, img):
-        raise NotImplementedError("override me!")
+        raise NotImplementedError
 
     def apply_transient_for(self, wid):
-        raise NotImplementedError("override me!")
+        raise NotImplementedError
 
     def paint_spinner(self, context, area):
-        raise NotImplementedError("override me!")
+        raise NotImplementedError
 
     def _pointer_modifiers(self, event):
-        raise NotImplementedError("override me!")
+        raise NotImplementedError
 
 
     def xget_u32_property(self, target, name):
-        raise NotImplementedError("override me!")
+        raise NotImplementedError
 
 
     def is_OR(self):
@@ -248,10 +254,10 @@ class ClientWindowBase(ClientWidgetBase):
 
     def _force_size_constraint(self, *size):
         return {
-            b"size-constraints" : {
-                b"maximum-size" : size,
-                b"minimum-size" : size,
-                b"base-size" : size,
+            "size-constraints" : {
+                "maximum-size" : size,
+                "minimum-size" : size,
+                "base-size" : size,
                 }
             }
 
@@ -262,11 +268,12 @@ class ClientWindowBase(ClientWidgetBase):
                 return title
             #perform metadata variable substitutions:
             #full of py3k unicode headaches that don't need to be
+            UNKNOWN_MACHINE = "<unknown machine>"
             default_values = {
                 "title"           : "<untitled window>",
-                "client-machine"  : "<unknown machine>",
+                "client-machine"  : UNKNOWN_MACHINE,
                 "windowid"        : str(self._id),
-                "server-machine"  : getattr(self._client, "_remote_hostname", None) or "<unknown machine>",
+                "server-machine"  : getattr(self._client, "_remote_hostname", None) or UNKNOWN_MACHINE,
                 "server-display"  : getattr(self._client, "_remote_display", None) or "<unknown display>",
                 }
             def getvar(var):
@@ -291,18 +298,14 @@ class ClientWindowBase(ClientWidgetBase):
                         if value not in (
                             "localhost",
                             "localhost.localdomain",
-                            "<unknown machine>",
                             "",
                             None):
                             return value
-                    return "<unknown machine>"
-                value = metadata.bytesget(var) or self._metadata.bytesget(var)
+                    return UNKNOWN_MACHINE
+                value = metadata.get(var) or self._metadata.get(var)
                 if value is None:
                     return default_values.get(var, "<unknown %s>" % var)
-                try:
-                    return value.decode("utf-8")
-                except UnicodeDecodeError:
-                    return str(value)
+                return net_utf8(value)
             def metadata_replace(match):
                 atvar = match.group(0)          #ie: '@title@'
                 var = atvar[1:len(atvar)-1]     #ie: 'title'
@@ -326,53 +329,53 @@ class ClientWindowBase(ClientWidgetBase):
             metalog.info("set_metadata: %s=%s", x, metadata.get(x))
         #WARNING: "class-instance" needs to go first because others may realize the window
         #(and GTK doesn't set the "class-instance" once the window is realized)
-        if b"class-instance" in metadata:
+        if "class-instance" in metadata:
             self.set_class_instance(*self._metadata.strtupleget("class-instance", ("xpra", "Xpra"), 2, 2))
             self.reset_icon()
 
-        if b"title" in metadata:
+        if "title" in metadata:
             title = self._get_window_title(metadata)
             self.set_title(title)
 
-        if b"icon-title" in metadata:
-            icon_title = metadata.strget("icon-title")
-            self.set_icon_name(icon_title)
+        if "icon-title" in metadata:
+            icon_title = metadata.strget("icon-title", "")
+            self.set_icon_name(net_utf8(icon_title))
             #the DE may have reset the icon now,
             #force it to use the one we really want:
             self.reset_icon()
 
-        if b"size-constraints" in metadata:
+        if "size-constraints" in metadata:
             sc = typedict(metadata.dictget("size-constraints", {}))
             self.size_constraints = sc
             self._set_initial_position = sc.boolget("set-initial-position", self._set_initial_position)
             self.set_size_constraints(sc, self.max_window_size)
 
-        if b"set-initial-position" in metadata:
+        if "set-initial-position" in metadata:
             #this should be redundant - but we keep it here for consistency
             self._set_initial_position = metadata.boolget("set-initial-position")
 
-        if b"transient-for" in metadata:
+        if "transient-for" in metadata:
             wid = metadata.intget("transient-for", -1)
             self.apply_transient_for(wid)
 
-        if b"modal" in metadata:
+        if "modal" in metadata:
             modal = metadata.boolget("modal")
             self.set_modal(modal)
 
         #apply window-type hint if window has not been mapped yet:
-        if b"window-type" in metadata and not self.get_mapped():
+        if "window-type" in metadata and not self.get_mapped():
             window_types = metadata.strtupleget("window-type")
             self.set_window_type(window_types)
 
-        if b"role" in metadata:
+        if "role" in metadata:
             role = metadata.strget("role")
             self.set_role(role)
 
-        if b"xid" in metadata:
+        if "xid" in metadata:
             xid = metadata.strget("xid")
             self.set_xid(xid)
 
-        if b"opacity" in metadata:
+        if "opacity" in metadata:
             opacity = metadata.intget("opacity", -1)
             if opacity<0:
                 opacity = 1
@@ -382,7 +385,7 @@ class ClientWindowBase(ClientWidgetBase):
             if hasattr(self, "set_opacity"):
                 self.set_opacity(opacity)
 
-        if b"has-alpha" in metadata:
+        if "has-alpha" in metadata:
             new_alpha = metadata.boolget("has-alpha")
             if new_alpha!=self._has_alpha:
                 l = alphalog
@@ -394,7 +397,7 @@ class ClientWindowBase(ClientWidgetBase):
                 l(" from %s to %s, behaviour is undefined", self._has_alpha, new_alpha)
                 self._has_alpha = new_alpha
 
-        if b"maximized" in metadata:
+        if "maximized" in metadata:
             maximized = metadata.boolget("maximized")
             if maximized!=self._maximized:
                 self._maximized = maximized
@@ -403,13 +406,13 @@ class ClientWindowBase(ClientWidgetBase):
                 else:
                     self.unmaximize()
 
-        if b"fullscreen" in metadata:
+        if "fullscreen" in metadata:
             fullscreen = metadata.boolget("fullscreen")
             if self._fullscreen is None or self._fullscreen!=fullscreen:
                 self._fullscreen = fullscreen
                 self.set_fullscreen(fullscreen)
 
-        if b"iconic" in metadata:
+        if "iconic" in metadata:
             iconified = metadata.boolget("iconic")
             if self._iconified!=iconified:
                 self._iconified = iconified
@@ -418,7 +421,7 @@ class ClientWindowBase(ClientWidgetBase):
                 else:
                     self.deiconify()
 
-        if b"decorations" in metadata:
+        if "decorations" in metadata:
             decorated = metadata.boolget("decorations", True)
             was_decorated = self.get_decorated()
             if WIN32 and decorated!=was_decorated:
@@ -428,25 +431,25 @@ class ClientWindowBase(ClientWidgetBase):
                 self.set_decorated(metadata.boolget("decorations"))
                 self.apply_geometry_hints(self.geometry_hints)
 
-        if b"above" in metadata:
+        if "above" in metadata:
             above = metadata.boolget("above")
             if self._above!=above:
                 self._above = above
                 self.set_keep_above(above)
 
-        if b"below" in metadata:
+        if "below" in metadata:
             below = metadata.boolget("below")
             if self._below!=below:
                 self._below = below
                 self.set_keep_below(below)
 
-        if b"shaded" in metadata:
+        if "shaded" in metadata:
             shaded = metadata.boolget("shaded")
             if self._shaded!=shaded:
                 self._shaded = shaded
                 self.set_shaded(shaded)
 
-        if b"sticky" in metadata:
+        if "sticky" in metadata:
             sticky = metadata.boolget("sticky")
             if self._sticky!=sticky:
                 self._sticky = sticky
@@ -455,54 +458,60 @@ class ClientWindowBase(ClientWidgetBase):
                 else:
                     self.unstick()
 
-        if b"skip-taskbar" in metadata:
+        if "skip-taskbar" in metadata:
             skip_taskbar = metadata.boolget("skip-taskbar")
             if self._skip_taskbar!=skip_taskbar:
                 self._skip_taskbar = skip_taskbar
                 self.set_skip_taskbar_hint(skip_taskbar)
 
-        if b"skip-pager" in metadata:
+        if "skip-pager" in metadata:
             skip_pager = metadata.boolget("skip-pager")
             if self._skip_pager!=skip_pager:
                 self._skip_pager = skip_pager
                 self.set_skip_pager_hint(skip_pager)
 
-        if b"workspace" in metadata:
+        if "opaque-region" in metadata:
+            opaque_region = metadata.tupleget("opaque-region", None, 0)
+            if self._opaque_region!=opaque_region:
+                self._opaque_region = opaque_region
+                self.set_opaque_region(opaque_region)
+
+        if "workspace" in metadata:
             self.set_workspace(metadata.intget("workspace"))
 
-        if b"bypass-compositor" in metadata:
+        if "bypass-compositor" in metadata:
             self.set_bypass_compositor(metadata.intget("bypass-compositor"))
 
-        if b"strut" in metadata:
+        if "strut" in metadata:
             self.set_strut(metadata.dictget("strut", {}))
 
-        if b"fullscreen-monitors" in metadata:
+        if "fullscreen-monitors" in metadata:
             self.set_fullscreen_monitors(metadata.inttupleget("fullscreen-monitors"))
 
-        if b"shape" in metadata:
+        if "shape" in metadata:
             self.set_shape(metadata.dictget("shape", {}))
 
-        if b"command" in metadata:
+        if "command" in metadata:
             self.set_command(metadata.strget("command"))
 
-        if b"x11-property" in metadata:
+        if "x11-property" in metadata:
             self.set_x11_property(*metadata.tupleget("x11-property"))
 
-        if b"content-type" in metadata:
+        if "content-type" in metadata:
             self.content_type = metadata.strget("content-type")
 
 
     def set_x11_property(self, *x11_property):
-        pass
+        pass        #see gtk client window base
 
     def set_command(self, command):
-        pass
+        pass        #see gtk client window base
 
     def set_class_instance(self, wmclass_name, wmclass_class):
-        pass
+        pass        #see gtk client window base
 
     def set_shape(self, shape):
-        log("set_shape(%s) not implemented", shape)
+        log("set_shape(%s) not implemented by %s", shape, type(self))
 
     def set_bypass_compositor(self, v):
         pass        #see gtk client window base
@@ -527,21 +536,21 @@ class ClientWindowBase(ClientWidgetBase):
         hints = typedict()
         client = self._client
         for (a, h1, h2) in (
-            (b"maximum-size", b"max_width", b"max_height"),
-            (b"minimum-size", b"min_width", b"min_height"),
-            (b"base-size", b"base_width", b"base_height"),
-            (b"increment", b"width_inc", b"height_inc"),
+            ("maximum-size", "max_width", "max_height"),
+            ("minimum-size", "min_width", "min_height"),
+            ("base-size", "base_width", "base_height"),
+            ("increment", "width_inc", "height_inc"),
             ):
             v = size_constraints.intpair(a)
             geomlog("intpair(%s)=%s", a, v)
             if v:
                 v1, v2 = v
-                if a==b"maximum-size" and v1>=32000 and v2>=32000 and WIN32:
+                if a=="maximum-size" and v1>=32000 and v2>=32000 and WIN32:
                     #causes problems, see #2714
                     continue
                 sv1 = client.sx(v1)
                 sv2 = client.sy(v2)
-                if a in (b"base-size", b"increment"):
+                if a in ("base-size", "increment"):
                     #rounding is not allowed for these values
                     fsv1 = client.fsx(v1)
                     fsv2 = client.fsy(v2)
@@ -558,8 +567,8 @@ class ClientWindowBase(ClientWidgetBase):
                 hints[h1], hints[h2] = sv1, sv2
         if not OSX:
             for (a, h) in (
-                (b"minimum-aspect-ratio", b"min_aspect"),
-                (b"maximum-aspect-ratio", b"max_aspect"),
+                ("minimum-aspect-ratio", "min_aspect"),
+                ("maximum-aspect-ratio", "max_aspect"),
                 ):
                 v = size_constraints.intpair(a)
                 if v:
@@ -569,8 +578,8 @@ class ClientWindowBase(ClientWidgetBase):
         w,h = max_window_size
         if w>0 and h>0 and not self._fullscreen:
             #get the min size, if there is one:
-            minw = max(1, hints.intget(b"min_width", 1))
-            minh = max(1, hints.intget(b"min_height", 1))
+            minw = max(1, hints.intget("min_width", 1))
+            minh = max(1, hints.intget("min_height", 1))
             #the actual max size is:
             # * greater than the min-size
             # * the lowest of the max-size set by the application and the one we have
@@ -578,20 +587,21 @@ class ClientWindowBase(ClientWidgetBase):
             #according to the GTK docs:
             #allowed window widths are base_width + width_inc * N where N is any integer
             #allowed window heights are base_height + width_inc * N where N is any integer
-            maxw = hints.intget(b"max_width", 32768)
-            maxh = hints.intget(b"max_height", 32768)
+            maxw = hints.intget("max_width", 32768)
+            maxh = hints.intget("max_height", 32768)
             maxw = max(minw, min(w, maxw))
             maxh = max(minh, min(h, maxh))
-            rw = (maxw - hints.intget(b"base_width", 0)) % max(hints.intget(b"width_inc", 1), 1)
-            rh = (maxh - hints.intget(b"base_height", 0)) % max(hints.intget(b"height_inc", 1), 1)
+            rw = (maxw - hints.intget("base_width", 0)) % max(hints.intget("width_inc", 1), 1)
+            rh = (maxh - hints.intget("base_height", 0)) % max(hints.intget("height_inc", 1), 1)
             maxw -= rw
             maxh -= rh
             #if the hints combination is invalid, it's possible that we'll end up
             #not honouring "base" + "inc", but honouring just "min" instead:
             maxw = max(minw, maxw)
             maxh = max(minh, maxh)
-            geomlog("modified hints for max window size %s: %s (rw=%s, rh=%s) -> max=%sx%s",
-                    max_window_size, hints, rw, rh, maxw, maxh)
+            scaling_info = self._client.xscale, self._client.yscale
+            geomlog("modified hints for max window size %s and scaling %s: %s (rw=%s, rh=%s) -> max=%sx%s",
+                    max_window_size, scaling_info, hints, rw, rh, maxw, maxh)
             #ensure we don't have duplicates with bytes / strings,
             #and that keys are always "bytes":
             #(in practice this code should never fire, just here as a reminder)
@@ -600,8 +610,8 @@ class ClientWindowBase(ClientWidgetBase):
             #bug 2214: GTK3 on win32 gets confused if we specify a large max-size
             # and it will mess up maximizing the window
             if not WIN32 or (maxw<32000 or maxh<32000):
-                hints[b"max_width"] = maxw
-                hints[b"max_height"] = maxh
+                hints["max_width"] = maxw
+                hints["max_height"] = maxh
         try:
             geomlog("calling: %s(%s)", self.apply_geometry_hints, hints)
             #save them so the window hooks can use the last value used:
@@ -622,16 +632,16 @@ class ClientWindowBase(ClientWidgetBase):
 
 
     def set_window_type(self, window_types):
-        pass
+        pass        #see gtk client window base
 
     def set_workspace(self, workspace):
-        pass
+        pass        #see gtk client window base
 
     def set_fullscreen(self, fullscreen):
-        pass
+        pass        #see gtk client window base
 
     def set_xid(self, xid):
-        pass
+        pass        #see gtk client window base
 
 
     def toggle_debug(self, *_args):
@@ -709,7 +719,7 @@ class ClientWindowBase(ClientWidgetBase):
 
     def repaint(self, x, y, w, h):
         #self.queue_draw_area(0, 0, *self._size)
-        raise NotImplementedError("no repaint on %s", type(self))
+        raise NotImplementedError("no repaint on %s" % type(self))
 
     def refresh_window(self, *args):
         log("refresh_window(%s) wid=%s", args, self._id)
@@ -726,14 +736,13 @@ class ClientWindowBase(ClientWidgetBase):
         backing = self._backing
         if not backing:
             log("draw_region: window %s has no backing, gone?", self._id)
-            from xpra.client.window_backing_base import fire_paint_callbacks
             fire_paint_callbacks(callbacks, -1, "no backing")
             return
         #only register this callback if we actually need it:
         if backing.draw_needs_refresh:
             if not backing.repaint_all:
                 self.pending_refresh.append((x, y, width, height))
-            if options.intget("flush", 0)==0:
+            if options.intget("flush", 0)==0 or FORCE_FLUSH:
                 callbacks.append(self.after_draw_refresh)
         if coding=="void":
             fire_paint_callbacks(callbacks)
@@ -800,7 +809,10 @@ class ClientWindowBase(ClientWidgetBase):
         self._client.quit(0)
 
     def void(self):
-        pass
+        """
+        This method can be used to capture key shortcuts
+        without triggering any specific action.
+        """
 
     def show_window_info(self, *args):
         from xpra.client.gtk_base.window_info import WindowInfo
@@ -843,20 +855,18 @@ class ClientWindowBase(ClientWidgetBase):
         self._client.rpc_call("dbus", rpc_args, **kwargs)
 
 
-    def get_mouse_event_wid(self, _x, _y):
-        #overriden in GTKClientWindowBase
+    def get_mouse_event_wid(self, *_args):
+        #used to be overriden in GTKClientWindowBase
         return self._id
 
     def _do_motion_notify_event(self, event):
         if self._client.readonly or self._client.server_readonly or not self._client.server_pointer:
             return
-        pointer, relative_pointer, modifiers, buttons = self._pointer_modifiers(event)
-        wid = self.get_mouse_event_wid(*pointer)
-        mouselog("do_motion_notify_event(%s) wid=%s / focus=%s / window wid=%i, device=%s, pointer=%s, relative pointer=%s, modifiers=%s, buttons=%s", event, wid, self._client._focused, self._id, self._device_info(event), pointer, relative_pointer, modifiers, buttons)
-        pdata = pointer
-        if self._client.server_pointer_relative:
-            pdata = list(pointer)+list(relative_pointer)
-        packet = ["pointer-position", wid, pdata, modifiers, buttons]
+        pointer_data, modifiers, buttons = self._pointer_modifiers(event)
+        wid = self.get_mouse_event_wid(*pointer_data)
+        mouselog("do_motion_notify_event(%s) wid=%s / focus=%s / window wid=%i, device=%s, pointer=%s, modifiers=%s, buttons=%s",
+                 event, wid, self._client._focused, self._id, self._device_info(event), pointer_data, modifiers, buttons)
+        packet = ("pointer-position", wid, pointer_data, modifiers, buttons)
         self._client.send_mouse_position(packet)
 
     def _device_info(self, event):
@@ -868,10 +878,10 @@ class ClientWindowBase(ClientWidgetBase):
     def _button_action(self, button, event, depressed, *args):
         if self._client.readonly or self._client.server_readonly or not self._client.server_pointer:
             return
-        pointer, relative_pointer, modifiers, buttons = self._pointer_modifiers(event)
-        wid = self.get_mouse_event_wid(*pointer)
+        pointer_data, modifiers, buttons = self._pointer_modifiers(event)
+        wid = self.get_mouse_event_wid(*pointer_data)
         mouselog("_button_action(%s, %s, %s) wid=%s / focus=%s / window wid=%i, device=%s, pointer=%s, modifiers=%s, buttons=%s",
-                 button, event, depressed, wid, self._client._focused, self._id, self._device_info(event), pointer, modifiers, buttons)
+                 button, event, depressed, wid, self._client._focused, self._id, self._device_info(event), pointer_data, modifiers, buttons)
         #map wheel buttons via translation table to support inverted axes:
         server_button = button
         if button>3:
@@ -886,14 +896,11 @@ class ClientWindowBase(ClientWidgetBase):
                     continue
                 b = sb
             server_buttons.append(b)
-        pdata = pointer
-        if self._client.server_pointer_relative:
-            pdata = list(pointer)+list(relative_pointer)
         def send_button(pressed):
-            self._client.send_button(wid, server_button, pressed, pdata, modifiers, server_buttons, *args)
+            self._client.send_button(wid, server_button, pressed, pointer_data, modifiers, server_buttons, *args)
         pressed_state = self.button_state.get(button, False)
         if SIMULATE_MOUSE_DOWN and pressed_state is False and depressed is False:
-            mouselog("button action: simulating a missing mouse-down event for window %s before sending the mouse-up event", wid)
+            mouselog("button action: simulating missing mouse-down event for window %s before mouse-up", wid)
             #(needed for some dialogs on win32):
             send_button(True)
         self.button_state[button] = depressed
