@@ -1,12 +1,13 @@
 # This file is part of Xpra.
 # Copyright (C) 2013 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2012-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2012-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 import time
-from ctypes import c_char_p
+from time import monotonic
+from gi.repository import GLib
 
 from OpenGL import version as OpenGL_version
 from OpenGL.error import GLError
@@ -51,7 +52,7 @@ from OpenGL.GL.ARB.framebuffer_object import (
     )
 
 from xpra.os_util import (
-    monotonic_time, strtobytes, hexstr,
+    strtobytes, bytestostr, hexstr,
     POSIX,
     DummyContextManager,
     )
@@ -208,7 +209,8 @@ TEX_RGB = 3
 TEX_FBO = 4         #FBO texture (guaranteed up-to-date window contents)
 TEX_TMP_FBO = 5
 TEX_CURSOR = 6
-N_TEXTURES = 7
+TEX_FPS = 7
+N_TEXTURES = 8
 
 # Shader number assignment
 YUV2RGB_SHADER = 0
@@ -248,7 +250,7 @@ class GLWindowBackingBase(WindowBackingBase):
         self.offscreen_fbo = None
         self.tmp_fbo = None
         self.pending_fbo_paint = []
-        self.last_flush = monotonic_time()
+        self.last_flush = monotonic()
         self.last_present_fbo_error = None
 
         super().__init__(wid, window_alpha and self.HAS_ALPHA)
@@ -265,12 +267,14 @@ class GLWindowBackingBase(WindowBackingBase):
 
     def get_info(self):
         info = super().get_info()
+        tpf = self.texture_pixel_format
+        tif = self.internal_format
         info.update({
             "type"                  : "OpenGL",
             "bit-depth"             : self.bit_depth,
             "pixel-format"          : self.pixel_format,
-            "texture-pixel-format"  : CONSTANT_TO_PIXEL_FORMAT.get(self.texture_pixel_format, str(self.texture_pixel_format)),
-            "internal-format"       : INTERNAL_FORMAT_TO_STR.get(self.internal_format, str(self.internal_format)),
+            "texture-pixel-format"  : CONSTANT_TO_PIXEL_FORMAT.get(tpf) or str(tpf),
+            "internal-format"       : INTERNAL_FORMAT_TO_STR.get(tif) or str(tif),
             })
         return info
 
@@ -335,7 +339,7 @@ class GLWindowBackingBase(WindowBackingBase):
             self.RGB_MODES)
 
     def get_encoding_properties(self):
-        props = WindowBackingBase.get_encoding_properties(self)
+        props = super().get_encoding_properties()
         if SCROLL_ENCODING:
             props["encoding.scrolling"] = True
         props["encoding.bit-depth"] = self.bit_depth
@@ -350,6 +354,7 @@ class GLWindowBackingBase(WindowBackingBase):
         #(see gl_init)
         self.render_size = ww, wh
         if self.size!=(bw, bh):
+            self.cancel_fps_refresh()
             self.gl_setup = False
             oldw, oldh = self.size
             self.size = bw, bh
@@ -398,7 +403,6 @@ class GLWindowBackingBase(WindowBackingBase):
                 with context:
                     self.pending_fbo_paint = ((0, 0, bw, bh), )
                     self.do_present_fbo()
-            from gi.repository import GLib
             GLib.timeout_add(FBO_RESIZE_DELAY, redraw)
 
 
@@ -407,6 +411,7 @@ class GLWindowBackingBase(WindowBackingBase):
         if not bool(glStringMarkerGREMEDY):
             return
         s = str(msg)
+        from ctypes import c_char_p
         c_string = c_char_p(s)
         glStringMarkerGREMEDY(0, c_string)
 
@@ -568,7 +573,10 @@ class GLWindowBackingBase(WindowBackingBase):
 
 
     def close_gl_config(self):
-        pass
+        """
+        Subclasses may free up resources at this point.
+        The GTK3 GL drawing area does.
+        """
 
     def close(self):
         self.close_gl_config()
@@ -584,6 +592,7 @@ class GLWindowBackingBase(WindowBackingBase):
         if b:
             self._backing = None
             b.destroy()
+        super().close()
 
 
     def paint_scroll(self, scroll_data, options, callbacks):    #pylint: disable=arguments-differ
@@ -700,14 +709,18 @@ class GLWindowBackingBase(WindowBackingBase):
             return
         #flush>0 means we should wait for the final flush=0 paint
         if flush==0 or not PAINT_FLUSH:
-            try:
-                with paint_context_manager:
-                    self.do_present_fbo()
-            except Exception as e:
-                log.error("Error presenting FBO:")
-                log.error(" %s", e)
-                log("Error presenting FBO", exc_info=True)
-                self.last_present_fbo_error = str(e)
+            self.record_fps_event()
+            self.managed_present_fbo()
+
+    def managed_present_fbo(self):
+        try:
+            with paint_context_manager:
+                self.do_present_fbo()
+        except Exception as e:
+            log.error("Error presenting FBO:")
+            log.error(" %s", e)
+            log("Error presenting FBO", exc_info=True)
+            self.last_present_fbo_error = str(e)
 
     def do_present_fbo(self):
         bw, bh = self.size
@@ -784,6 +797,9 @@ class GLWindowBackingBase(WindowBackingBase):
         if self.border and self.border.shown:
             self.draw_border()
 
+        if self.is_show_fps():
+            self.draw_fps()
+
         # Show the backbuffer on screen
         glFlush()
         self.gl_show(rect_count)
@@ -832,7 +848,7 @@ class GLWindowBackingBase(WindowBackingBase):
 
     def draw_pointer(self):
         px, py, _, _, size, start_time = self.pointer_overlay
-        elapsed = monotonic_time()-start_time
+        elapsed = monotonic()-start_time
         log("pointer_overlay=%s, elapsed=%.1f, timeout=%s, cursor-data=%s",
             self.pointer_overlay, elapsed, CURSOR_IDLE_TIMEOUT, (self.cursor_data or [])[:7])
         if elapsed>=CURSOR_IDLE_TIMEOUT:
@@ -855,17 +871,20 @@ class GLWindowBackingBase(WindowBackingBase):
             glEnd()
             return
 
-        cw = self.cursor_data[3]
-        ch = self.cursor_data[4]
+        w = self.cursor_data[3]
+        h = self.cursor_data[4]
         xhot = self.cursor_data[5]
         yhot = self.cursor_data[6]
         x = px-xhot
         y = py-yhot
-        #paint the texture containing the cursor:
+        self.blend_texture(self.textures[TEX_CURSOR], x, y, w, h)
+
+    def blend_texture(self, texture, x, y, w, h):
+        #paint this texture
         glActiveTexture(GL_TEXTURE0)
         target = GL_TEXTURE_RECTANGLE_ARB
         glEnable(target)
-        glBindTexture(target, self.textures[TEX_CURSOR])
+        glBindTexture(target, texture)
         glEnable(GL_BLEND)
         glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA)
         #glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)
@@ -873,12 +892,12 @@ class GLWindowBackingBase(WindowBackingBase):
         glBegin(GL_QUADS)
         glTexCoord2i(0, 0)
         glVertex2i(x, y)
-        glTexCoord2i(0, ch)
-        glVertex2i(x, y+ch)
-        glTexCoord2i(cw, ch)
-        glVertex2i(x+cw, y+ch)
-        glTexCoord2i(cw, 0)
-        glVertex2i(x+cw, y)
+        glTexCoord2i(0, h)
+        glVertex2i(x, y+h)
+        glTexCoord2i(w, h)
+        glVertex2i(x+w, y+h)
+        glTexCoord2i(w, 0)
+        glVertex2i(x+w, y)
         glEnd()
 
         glDisable(GL_BLEND)
@@ -900,6 +919,27 @@ class GLWindowBackingBase(WindowBackingBase):
             glVertex2i(px, py)
         glEnd()
 
+    def update_fps_buffer(self, width, height, pixels):
+        #we always call 'record_fps_event' from a gl context,
+        #so it is safe to upload the texture:
+        self.upload_rgba_texture(self.textures[TEX_FPS], width, height, pixels)
+
+    def draw_fps(self):
+        x, y = 2, 5
+        width, height = self.fps_buffer_size
+        self.blend_texture(self.textures[TEX_FPS], x, y, width, height)
+        self.cancel_fps_refresh()
+        def refresh_screen():
+            self.fps_refresh_timer = 0
+            log("refresh_screen()")
+            if not self.paint_screen:
+                return
+            context = self.gl_context()
+            with context:
+                self.update_fps()
+                self.managed_present_fbo()
+        self.fps_refresh_timer = GLib.timeout_add(1000, refresh_screen)
+
 
     def validate_cursor(self):
         cursor_data = self.cursor_data
@@ -932,22 +972,22 @@ class GLWindowBackingBase(WindowBackingBase):
             return
         with context:
             self.gl_init()
-            self.upload_cursor_texture(cw, ch, pixels)
+            self.upload_rgba_texture(self.textures[TEX_CURSOR], cw, ch, pixels)
 
-    def upload_cursor_texture(self, width : int, height : int, pixels):
+    def upload_rgba_texture(self, texture, width : int, height : int, pixels):
         upload, pixel_data = self.pixels_for_upload(pixels)
         rgb_format = "RGBA"
         glActiveTexture(GL_TEXTURE0)
         target = GL_TEXTURE_RECTANGLE_ARB
         glEnable(target)
-        glBindTexture(target, self.textures[TEX_CURSOR])
+        glBindTexture(target, texture)
         self.set_alignment(width, width*4, rgb_format)
         glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
         glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
         glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER)
         glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER)
         glTexImage2D(target, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixel_data)
-        log("GL cursor %ix%i uploaded %i bytes of %s pixel data using %s",
+        log("upload_rgba_texture %ix%i uploaded %i bytes of %s pixel data using %s",
             width, height, len(pixels), rgb_format, upload)
         glBindTexture(target, 0)
         glDisable(target)
@@ -1013,7 +1053,10 @@ class GLWindowBackingBase(WindowBackingBase):
                           x, y, w, h, width, height, options, callbacks)
         else:
             img = self.jpeg_decoder.decompress_to_rgb("BGRX", img_data)
-            self.idle_add(self.do_paint_rgb, "BGRX", img.get_pixels(), x, y, width, height, img.get_rowstride(), options, callbacks)
+            w = img.get_width()
+            h = img.get_height()
+            self.idle_add(self.do_paint_rgb, "BGRX", img.get_pixels(), x, y, w, h, width, height,
+                          img.get_rowstride(), options, callbacks)
 
     def paint_webp(self, img_data, x : int, y : int, width : int, height : int, options, callbacks):
         subsampling = options.strget("subsampling")
@@ -1042,10 +1085,7 @@ class GLWindowBackingBase(WindowBackingBase):
         if not options.boolget("paint", True):
             fire_paint_callbacks(callbacks)
             return
-        try:
-            rgb_format = rgb_format.decode("utf-8")
-        except (AttributeError, UnicodeDecodeError):
-            pass
+        rgb_format = bytestostr(rgb_format)
         try:
             upload, img_data = self.pixels_for_upload(img_data)
 
@@ -1107,7 +1147,10 @@ class GLWindowBackingBase(WindowBackingBase):
         fire_paint_callbacks(callbacks, False, message)
 
 
-    def do_video_paint(self, img, x : int, y : int, enc_width : int, enc_height : int, width : int, height : int, options, callbacks):
+    def do_video_paint(self, img,
+                       x : int, y : int, enc_width : int, enc_height : int, width : int, height : int,
+                       options, callbacks):
+        log("do_video_paint%s", (x, y, enc_width, enc_height, width, height, options, callbacks))
         if not zerocopy_upload or FORCE_CLONE:
             #copy so the data will be usable (usually a str)
             img.clone_pixel_data()
@@ -1115,7 +1158,8 @@ class GLWindowBackingBase(WindowBackingBase):
         if pixel_format in ("GBRP10", "YUV444P10"):
             #call superclass to handle csc
             #which will end up calling paint rgb with r210 data
-            return super().do_video_paint(img, x, y, enc_width, enc_height, width, height, options, callbacks)
+            super().do_video_paint(img, x, y, enc_width, enc_height, width, height, options, callbacks)
+            return
         if pixel_format.startswith("GBRP"):
             shader = RGBP2RGB_SHADER
         else:
@@ -1260,9 +1304,9 @@ class GLWindowBackingBase(WindowBackingBase):
 
 
     def gl_show(self, rect_count):
-        start = monotonic_time()
+        start = monotonic()
         self.do_gl_show(rect_count)
-        end = monotonic_time()
+        end = monotonic()
         flush_elapsed = end-self.last_flush
         self.last_flush = end
         fpslog("gl_show after %3ims took %2ims, %2i updates", flush_elapsed*1000, (end-start)*1000, rect_count)
