@@ -4,6 +4,7 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+from time import monotonic
 import cairo
 from cairo import (  #pylint: disable=no-name-in-module
     Context, ImageSurface,
@@ -14,12 +15,14 @@ from gi.repository import GLib, Gdk
 
 from xpra.client.paint_colors import get_paint_box_color
 from xpra.client.window_backing_base import WindowBackingBase, fire_paint_callbacks, SCROLL_ENCODING
-from xpra.os_util import memoryview_to_bytes, monotonic_time
+from xpra.os_util import memoryview_to_bytes
 from xpra.util import envbool
 from xpra.log import Logger
 
 log = Logger("paint", "cairo")
 
+COPY_OLD_BACKING = envbool("XPRA_CAIRO_COPY_OLD_BACKING", True)
+
 FORMATS = {-1   : "INVALID"}
 for attr in dir(cairo):
     if attr.startswith("FORMAT_"):
@@ -29,7 +32,7 @@ for attr in dir(cairo):
 def cairo_paint_pointer_overlay(context, cursor_data, px : int, py : int, start_time):
     if not cursor_data:
         return
-    elapsed = max(0, monotonic_time()-start_time)
+    elapsed = max(0, monotonic()-start_time)
     if elapsed>6:
         return
     from xpra.gtk_common.gtk_util import get_pixbuf_from_data
@@ -63,6 +66,7 @@ class CairoBackingBase(WindowBackingBase):
         self.idle_add = GLib.idle_add
         self.size = 0, 0
         self.render_size = 0, 0
+        self.fps_image = None
 
     def init(self, ww : int, wh : int, bw : int, bh : int):
         mod = self.size!=(bw, bh) or self.render_size!=(ww, wh)
@@ -91,16 +95,20 @@ class CairoBackingBase(WindowBackingBase):
         backing = ImageSurface(FORMAT_ARGB32, bw, bh)
         self._backing = backing
         cr = Context(backing)
-        cr.set_operator(OPERATOR_CLEAR)
-        cr.set_source_rgba(1, 1, 1, 1)
+        if self._alpha_enabled:
+            cr.set_operator(OPERATOR_CLEAR)
+            cr.set_source_rgba(1, 1, 1, 0)
+        else:
+            cr.set_operator(OPERATOR_SOURCE)
+            cr.set_source_rgba(1, 1, 1, 1)
         cr.rectangle(0, 0, bw, bh)
         cr.fill()
-        if old_backing is not None:
+        if COPY_OLD_BACKING and old_backing is not None:
             oldw, oldh = old_backing.get_width(), old_backing.get_height()
             sx, sy, dx, dy, w, h = self.gravity_copy_coords(oldw, oldh, bw, bh)
             cr.translate(dx-sx, dy-sy)
             cr.rectangle(sx, sy, w, h)
-            cr.fill()
+            cr.clip()
             cr.set_operator(OPERATOR_SOURCE)
             cr.set_source_surface(old_backing, 0, 0)
             cr.paint()
@@ -112,7 +120,7 @@ class CairoBackingBase(WindowBackingBase):
         if backing:
             backing.finish()
             self._backing = None
-        WindowBackingBase.close(self)
+        super().close()
 
 
     def cairo_paint_pixbuf(self, pixbuf, x : int, y : int, options):
@@ -159,9 +167,13 @@ class CairoBackingBase(WindowBackingBase):
 
         if self.paint_box_line_width:
             gc.restore()
-            encoding = options.get("encoding")
+            encoding = options.strget("encoding")
             self.cairo_paint_box(gc, encoding, x, y, width, height)
 
+        flush = options.intget("flush", 0)
+        if flush==0:
+            self.record_fps_event()
+
     def cairo_paint_box(self, gc, encoding, x, y, w, h):
         color = get_paint_box_color(encoding)
         gc.set_line_width(self.paint_box_line_width)
@@ -196,7 +208,7 @@ class CairoBackingBase(WindowBackingBase):
 
 
     def get_encoding_properties(self):
-        props = WindowBackingBase.get_encoding_properties(self)
+        props = super().get_encoding_properties()
         if SCROLL_ENCODING:
             props["encoding.scrolling"] = True
         return props
@@ -290,3 +302,19 @@ class CairoBackingBase(WindowBackingBase):
             spx = round(w*px/ww)
             spy = round(h*py/wh)
             cairo_paint_pointer_overlay(context, self.cursor_data, x+spx, y+spy, start_time)
+        if self.is_show_fps() and self.fps_image:
+            x, y = 10, 10
+            context.translate(x, y)
+            context.set_operator(OPERATOR_OVER)
+            context.set_source_surface(self.fps_image, 0, 0)
+            context.paint()
+            self.cancel_fps_refresh()
+            #width, height = self.fps_buffer_size
+            def refresh_screen():
+                self.fps_refresh_timer = 0
+                b = self._backing
+                if b:
+                    self.update_fps()
+                    w, h = self.render_size
+                    self.repaint(0, 0, w, h)
+            self.fps_refresh_timer = GLib.timeout_add(1000, refresh_screen)
