@@ -109,6 +109,7 @@ CLAMP_WINDOW_TO_SCREEN = envbool("XPRA_CLAMP_WINDOW_TO_SCREEN", True)
 FOCUS_RECHECK_DELAY = envint("XPRA_FOCUS_RECHECK_DELAY", 0)
 REPAINT_MAXIMIZED = envint("XPRA_REPAINT_MAXIMIZED", 0)
 REFRESH_MAXIMIZED = envbool("XPRA_REFRESH_MAXIMIZED", True)
+UNICODE_KEYNAMES = envbool("XPRA_UNICODE_KEYNAMES", False)
 
 WINDOW_OVERFLOW_TOP = envbool("XPRA_WINDOW_OVERFLOW_TOP", False)
 AWT_RECENTER = envbool("XPRA_AWT_RECENTER", True)
@@ -477,6 +478,7 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
         if self._focus_latest:
             self._focus()
         else:
+            self._client.pointer_grabbed = False
             self._unfocus()
 
     def _focus(self):
@@ -2145,11 +2147,21 @@ class GTKClientWindowBase(ClientWindowBase, Gtk.Window):
     def parse_key_event(self, event, pressed):
         keyval = event.keyval
         keycode = event.hardware_keycode
-        keyname = Gdk.keyval_name(keyval)
+        keyname = Gdk.keyval_name(keyval) or ""
         keyname = KEY_TRANSLATIONS.get((keyname, keyval, keycode), keyname)
+        if keyname.startswith("U+") and not UNICODE_KEYNAMES:
+            #workaround for MS Windows, try harder to find a valid key
+            #see ticket #3417
+            keymap = Gdk.Keymap.get_default()
+            r = keymap.get_entries_for_keycode(event.hardware_keycode)
+            if r[0]:
+                for kc in r[2]:
+                    keyname = Gdk.keyval_name(kc)
+                    if not keyname.startswith("U+"):
+                        break
         key_event = KeyEvent()
         key_event.modifiers = self._client.mask_to_names(event.state)
-        key_event.keyname = keyname or ""
+        key_event.keyname = keyname
         key_event.keyval = keyval or 0
         key_event.keycode = keycode
         key_event.group = event.group
