@@ -1,16 +1,24 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2011-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from gi.repository import Gtk
+import os
+import re
 
-from xpra.util import envbool
-from xpra.os_util import OSX
-from xpra.gtk_common.gtk_util import menuitem
+from gi.repository import Gtk, GdkPixbuf
+
+from xpra.util import envbool, repr_ellipsized
+from xpra.os_util import OSX, bytestostr
+from xpra.codecs.icon_util import INKSCAPE_RE
+from xpra.gtk_common.gtk_util import (
+    menuitem,
+    get_pixbuf_from_data, scaled_image,
+    )
 from xpra.gtk_common.about import about, close_about
 from xpra.platform.gui import get_icon_size
+from xpra.platform.paths import get_icon_dir
 from xpra.log import Logger
 
 log = Logger("menu")
@@ -19,6 +27,54 @@ MENU_ICONS = envbool("XPRA_MENU_ICONS", True)
 HIDE_DISABLED_MENU_ENTRIES = OSX
 
 
+LOSSLESS = "Lossless"
+QUALITY_OPTIONS_COMMON = {
+                50      : "Average",
+                30      : "Low",
+                }
+MIN_QUALITY_OPTIONS = QUALITY_OPTIONS_COMMON.copy()
+MIN_QUALITY_OPTIONS.update({
+    0 : "None",
+    75  : "High",
+    })
+MIN_QUALITY_OPTIONS = dict(sorted(MIN_QUALITY_OPTIONS.items()))
+QUALITY_OPTIONS = QUALITY_OPTIONS_COMMON.copy()
+QUALITY_OPTIONS.update({
+    0 : "Auto",
+    1   : "Lowest",
+    90  : "Best",
+    100 : LOSSLESS,
+    })
+QUALITY_OPTIONS = dict(sorted(QUALITY_OPTIONS.items()))
+
+
+SPEED_OPTIONS_COMMON = {
+                70      : "Low Latency",
+                50      : "Average",
+                30      : "Low Bandwidth",
+                }
+MIN_SPEED_OPTIONS = SPEED_OPTIONS_COMMON.copy()
+MIN_SPEED_OPTIONS[0] = "None"
+MIN_SPEED_OPTIONS = dict(sorted(MIN_SPEED_OPTIONS.items()))
+SPEED_OPTIONS = SPEED_OPTIONS_COMMON.copy()
+SPEED_OPTIONS.update({
+    0   : "Auto",
+    1   : "Lowest Bandwidth",
+    100 : "Lowest Latency",
+    })
+SPEED_OPTIONS = dict(sorted(SPEED_OPTIONS.items()))
+
+def get_bandwidth_menu_options():
+    options = []
+    for x in os.environ.get("XPRA_BANDWIDTH_MENU_OPTIONS", "1,2,5,10,20,50,100").split(","):
+        try:
+            options.append(int(float(x)*1000*1000))
+        except ValueError:
+            log.warn("Warning: invalid bandwidth menu option '%s'", x)
+    return options
+BANDWIDTH_MENU_OPTIONS = get_bandwidth_menu_options()
+
+
 def ll(m):
     try:
         return "%s:%s" % (type(m), m.get_label())
@@ -33,6 +89,58 @@ def set_sensitive(widget, sensitive):
             widget.hide()
     widget.set_sensitive(sensitive)
 
+def load_pixbuf(data):
+    loader = GdkPixbuf.PixbufLoader()
+    loader.write(data)
+    loader.close()
+    return loader.get_pixbuf()
+
+def get_appimage(app_name, icondata=None, menu_icon_size=24):
+    pixbuf = None
+    if app_name and not icondata:
+        #try to load from our icons:
+        nstr = bytestostr(app_name).lower()
+        icon_filename = os.path.join(get_icon_dir(), "%s.png" % nstr)
+        if os.path.exists(icon_filename):
+            pixbuf = GdkPixbuf.Pixbuf.new_from_file(icon_filename)
+    def err(e):
+        log("failed to load icon", exc_info=True)
+        log.error("Error: failed to load icon data for '%s':", bytestostr(app_name))
+        log.error(" %s", e)
+        log.error(" data=%s", repr_ellipsized(icondata))
+    if not pixbuf and icondata:
+        #gtk pixbuf loader:
+        try:
+            pixbuf = load_pixbuf(icondata)
+        except Exception as e:
+            log("pixbuf loader failed", exc_info=True)
+            if re.findall(INKSCAPE_RE, icondata):
+                try:
+                    pixbuf = load_pixbuf(re.sub(INKSCAPE_RE, b"", icondata))
+                    e = None
+                except Exception:
+                    #there is almost no chance pillow will be able to load it
+                    #(it doesn't even have svg support at time of writing)
+                    #so don't bother showing another error for the same data:
+                    icondata = None
+            if e:
+                err(e)
+    if not pixbuf and icondata:
+        #let's try pillow:
+        try:
+            from xpra.codecs.pillow.decoder import open_only  #pylint: disable=import-outside-toplevel
+            img = open_only(icondata)
+            has_alpha = img.mode=="RGBA"
+            width, height = img.size
+            rowstride = width * (3+int(has_alpha))
+            pixbuf = get_pixbuf_from_data(img.tobytes(), has_alpha, width, height, rowstride)
+            return scaled_image(pixbuf, icon_size=menu_icon_size)
+        except Exception:
+            err(e)
+    if pixbuf:
+        return scaled_image(pixbuf, icon_size=menu_icon_size)
+    return None
+
 
 #utility method to ensure there is always only one CheckMenuItem
 #selected in a submenu:
@@ -350,5 +458,27 @@ class MenuHelper:
             self.show_bug_report()
         return  self.menuitem("Bug Report", "bugs.png", None, show_bug_report_cb)
 
+    def make_docsmenuitem(self):
+        def show_docs(*_args):
+            from xpra.scripts.main import run_docs
+            from xpra.make_thread import start_thread
+            start_thread(run_docs, "open documentation", True)
+        return self.menuitem("Documentation", "documentation.png", None, show_docs)
+
+    def make_html5menuitem(self):
+        def show_html5(*_args):
+            from xpra.scripts.main import run_html5
+            from xpra.make_thread import start_thread
+            url_options = {}
+            try:
+                for k in ("port", "host", "username", "mode", "display"):
+                    v = self.client.display_desc.get(k)
+                    if v is not None:
+                        url_options[k] = v
+            except Exception:
+                pass
+            start_thread(run_html5, "open HTML5 client", True, args=(url_options, ))
+        return self.menuitem("HTML5 client", "browser.png", None, show_html5)
+
     def make_closemenuitem(self):
         return self.menuitem("Close Menu", "close.png", None, self.close_menu)
