@@ -6,11 +6,12 @@
 
 import os
 import re
+from time import monotonic
 from collections import deque
 
-from xpra.os_util import monotonic_time, POSIX
+from xpra.os_util import POSIX
 from xpra.util import envint, envbool, csv, typedict
-from xpra.exit_codes import EXIT_TIMEOUT
+from xpra.exit_codes import EXIT_CONNECTION_LOST
 from xpra.net.packet_encoding import ALL_ENCODERS
 from xpra.client.mixins.stub_client_mixin import StubClientMixin
 from xpra.scripts.config import parse_with_unit
@@ -27,6 +28,10 @@ LOG_INFO_RESPONSE = os.environ.get("XPRA_LOG_INFO_RESPONSE", "")
 AUTO_BANDWIDTH_PCT = envint("XPRA_AUTO_BANDWIDTH_PCT", 80)
 assert 1<AUTO_BANDWIDTH_PCT<=100, "invalid value for XPRA_AUTO_BANDWIDTH_PCT: %i" % AUTO_BANDWIDTH_PCT
 
+LOCAL_JITTER = envint("XPRA_LOCAL_JITTER", 0)
+WAN_JITTER = envint("XPRA_WAN_JITTER", 20)
+WIRELESS_JITTER = envint("XPRA_WIRELESS_JITTER", 1000)
+
 
 class NetworkState(StubClientMixin):
     """
@@ -36,7 +41,7 @@ class NetworkState(StubClientMixin):
     """
 
     def __init__(self):
-        StubClientMixin.__init__(self)
+        super().__init__()
         self.server_start_time = -1
         #legacy:
         self.compression_level = 0
@@ -127,21 +132,30 @@ class NetworkState(StubClientMixin):
         log("get_caps() found socket_speed=%s", socket_speed)
         if socket_speed:
             connection_data["speed"] = socket_speed
-        adapter_type = device_value("adapter-type")
-        log("get_caps() found adapter-type=%s", adapter_type)
-        if adapter_type:
-            connection_data["adapter-type"] = adapter_type
+        default_adapter_type = device_value("adapter-type")
+        log("get_caps() found default adapter-type=%s", default_adapter_type)
+        device_name = device_value("name")
+        log("get_caps() found device name=%s", device_name)
         jitter = device_value("jitter", int, -1)
         if jitter<0:
-            at = adapter_type.lower()
-            if any(at.find(x)>=0 for x in ("ether", "local", "fiber", "1394", "infiniband")):
-                jitter = 0
+            at = default_adapter_type.lower()
+            if device_name.startswith("wlan") or device_name.startswith("wlp") or device_name.find("wifi")>=0:
+                jitter = WIRELESS_JITTER
+                device_info["adapter-type"] = "wireless"
+            elif device_name=="lo":
+                jitter = LOCAL_JITTER
+                device_info["adapter-type"] = "loopback"
+            elif any(at.find(x)>=0 for x in ("ether", "local", "fiber", "1394", "infiniband")):
+                jitter = LOCAL_JITTER
             elif at.find("wan")>=0:
-                jitter = 20
-            elif at.find("wireless")>=0 or at.find("wifi")>=0 or at.find("80211")>=0:
-                jitter = 1000
+                jitter = WAN_JITTER
+            elif at.find("wireless")>=0 or at.find("wlan")>=0 or at.find("wifi")>=0 or at.find("80211")>=0:
+                jitter = WIRELESS_JITTER
         if jitter>=0:
             connection_data["jitter"] = jitter
+        adapter_type = device_value("adapter-type", str, default_adapter_type)
+        if adapter_type:
+            connection_data["adapter-type"] = adapter_type
         log("get_caps() connection-data=%s", connection_data)
         caps["connection-data"] = connection_data
         bandwidth_limit = self.bandwidth_limit
@@ -222,7 +236,7 @@ class NetworkState(StubClientMixin):
             return False
         last = self._server_ok
         if FAKE_BROKEN_CONNECTION>0:
-            self._server_ok = (int(monotonic_time()) % FAKE_BROKEN_CONNECTION) <= (FAKE_BROKEN_CONNECTION//2)
+            self._server_ok = (int(monotonic()) % FAKE_BROKEN_CONNECTION) <= (FAKE_BROKEN_CONNECTION//2)
         else:
             self._server_ok = self.last_ping_echoed_time>=ping_sent_time
         if not self._server_ok:
@@ -251,10 +265,14 @@ class NetworkState(StubClientMixin):
         log("check_echo_timeout(%s) last_ping_echoed_time=%s", ping_time, self.last_ping_echoed_time)
         if self.last_ping_echoed_time<ping_time:
             #no point trying to use disconnect_and_quit() to tell the server here..
-            self.warn_and_quit(EXIT_TIMEOUT, "server ping timeout - waited %s seconds without a response" % PING_TIMEOUT)
+            self.warn_and_quit(EXIT_CONNECTION_LOST, "server ping timeout - waited %s seconds without a response" % PING_TIMEOUT)
 
     def send_ping(self):
-        now_ms = int(1000.0*monotonic_time())
+        p = self._protocol
+        if not p or p.TYPE!="xpra":
+            self.ping_timer = None
+            return False
+        now_ms = int(1000.0*monotonic())
         self.send("ping", now_ms)
         wait = 2.0
         spl = tuple(self.server_ping_latency)
@@ -272,11 +290,11 @@ class NetworkState(StubClientMixin):
         echoedtime, l1, l2, l3, cl = packet[1:6]
         self.last_ping_echoed_time = echoedtime
         self.check_server_echo(0)
-        server_ping_latency = monotonic_time()-echoedtime/1000.0
-        self.server_ping_latency.append((monotonic_time(), server_ping_latency))
+        server_ping_latency = monotonic()-echoedtime/1000.0
+        self.server_ping_latency.append((monotonic(), server_ping_latency))
         self.server_load = l1, l2, l3
         if cl>=0:
-            self.client_ping_latency.append((monotonic_time(), cl/1000.0))
+            self.client_ping_latency.append((monotonic(), cl/1000.0))
         log("ping echo server load=%s, measured client latency=%sms", self.server_load, cl)
 
     def _process_ping(self, packet):
@@ -307,7 +325,8 @@ class NetworkState(StubClientMixin):
         self.send_deflate_level()
 
     def send_deflate_level(self):
-        if self._protocol:
+        p = self._protocol
+        if p and p.TYPE=="xpra":
             self._protocol.set_compression_level(self.compression_level)
             self.send("set_deflate", self.compression_level)
 
