@@ -13,13 +13,14 @@ import re
 
 from io import BytesIO
 
-from xpra.util import envint, first_time, ellipsizer
+from xpra.util import envint, envbool, first_time, ellipsizer
 from xpra.os_util import load_binary_file
 from xpra.log import Logger
 
 log = Logger("menu")
 
 MAX_ICON_SIZE = envint("XPRA_XDG_MAX_ICON_SIZE", 0)
+SVG_TO_PNG = envbool("XPRA_SVG_TO_PNG", True)
 
 INKSCAPE_RE = b'\\sinkscape:[a-zA-Z]*=["a-zA-Z0-9]*'
 
@@ -45,6 +46,9 @@ def load_Rsvg():
 
 
 def load_icon_from_file(filename, max_size=MAX_ICON_SIZE):
+    if os.path.isdir(filename):
+        log("load_icon_from_file(%s, %i) path is a directory!", filename, max_size)
+        return None
     log("load_icon_from_file(%s, %i)", filename, max_size)
     if filename.endswith("xpm"):
         from PIL import Image  #pylint: disable=import-outside-toplevel
@@ -80,7 +84,10 @@ def load_icon_from_file(filename, max_size=MAX_ICON_SIZE):
     return icondata, os.path.splitext(filename)[1].lstrip(".")
 
 def svg_to_png(filename, icondata, w=128, h=128):
+    if not SVG_TO_PNG:
+        return None
     Rsvg = load_Rsvg()
+    log("svg_to_png%s Rsvg=%s", (filename, "%i bytes" % len(icondata), w, h), Rsvg)
     if not Rsvg:
         return None
     try:
@@ -89,11 +96,16 @@ def svg_to_png(filename, icondata, w=128, h=128):
         img = ImageSurface(FORMAT_ARGB32, w, h)
         ctx = Context(img)
         handle = Rsvg.Handle.new_from_data(icondata)
+        dim = handle.get_dimensions()
+        ctx.scale(w/dim.width, h/dim.height)
         handle.render_cairo(ctx)
+        del handle
+        img.flush()
         buf = BytesIO()
         img.write_to_png(buf)
         icondata = buf.getvalue()
         buf.close()
+        img.finish()
         return icondata
     except Exception:
         log("svg_to_png%s", (icondata, w, h), exc_info=True)
@@ -106,3 +118,4 @@ def svg_to_png(filename, icondata, w=128, h=128):
         if filename:
             log.error(" '%s':", filename)
         log.error(" %i bytes, %s", len(icondata), ellipsizer(icondata))
+        return None
