@@ -1,9 +1,9 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 #
 #gtkPopupNotify.py
 #
 # Copyright 2009 Daniel Woodhouse
-# Copyright 2013-2018 Antoine Martin <antoine@xpra.org>
+# Copyright 2013-2021 Antoine Martin <antoine@xpra.org>
 #
 #This program is free software: you can redistribute it and/or modify
 #it under the terms of the GNU Lesser General Public License as published by
@@ -18,15 +18,16 @@
 #You should have received a copy of the GNU Lesser General Public License
 #along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-import os
 import gi
 gi.require_version("Gtk", "3.0")
 gi.require_version("Gdk", "3.0")
 from gi.repository import GLib, Gtk, Gdk, GdkPixbuf
 
-from xpra.os_util import OSX, bytestostr
+from xpra.os_util import OSX
+from xpra.util import u
 from xpra.gtk_common.gtk_util import (
     add_close_accel, color_parse,
+    get_icon_pixbuf,
     )
 from xpra.notifications.notifier_base import NotifierBase, log
 
@@ -40,14 +41,6 @@ DEFAULT_WIDTH = 340
 DEFAULT_HEIGHT = 100
 
 
-
-def get_pixbuf(icon_name):
-    from xpra.platform.paths import get_icon_dir
-    icon_filename = os.path.join(get_icon_dir(), icon_name)
-    if os.path.exists(icon_filename):
-        return GdkPixbuf.Pixbuf.new_from_file(icon_filename)
-    return None
-
 class GTK_Notifier(NotifierBase):
 
     def __init__(self, closed_cb=None, action_cb=None, size_x=DEFAULT_WIDTH, size_y=DEFAULT_HEIGHT, timeout=5):
@@ -102,7 +95,7 @@ class GTK_Notifier(NotifierBase):
         self._notify_stack = []
         for x in popups:
             x.hide_notification()
-        NotifierBase.cleanup(self)
+        super().cleanup()
 
 
     def get_origin_x(self):
@@ -120,7 +113,7 @@ class GTK_Notifier(NotifierBase):
     def show_notify(self, dbus_id, tray, nid,
                     app_name, replaces_nid, app_icon,
                     summary, body, actions, hints, timeout, icon):
-        self.new_popup(nid, summary, body, actions, icon, timeout, 0<timeout<=600)
+        GLib.idle_add(self.new_popup, nid, summary, body, actions, icon, timeout, 0<timeout<=600)
 
     def new_popup(self, nid, summary, body, actions, icon, timeout=10*1000, show_timeout=False):
         """Create a new Popup instance, or update an existing one """
@@ -142,7 +135,7 @@ class GTK_Notifier(NotifierBase):
         popup = Popup(self, nid, summary, body, actions, image=image, timeout=timeout//1000, show_timeout=show_timeout)
         self._notify_stack.append(popup)
         self._offset += self._notify_stack[-1].h
-        return popup
+        return False
 
     def destroy_popup_cb(self, popup):
         if popup in self._notify_stack:
@@ -168,7 +161,7 @@ class Popup(Gtk.Window):
         log("Popup%s", (stack, nid, title, message, actions, image, timeout, show_timeout))
         self.stack = stack
         self.nid = nid
-        Gtk.Window.__init__(self)
+        super().__init__()
 
         self.set_accept_focus(False)
         self.set_focus_on_map(False)
@@ -191,7 +184,7 @@ class Popup(Gtk.Window):
         self.header.set_padding(3, 3)
         self.header.set_alignment(0, 0)
         header_box.pack_start(self.header, True, True, 5)
-        icon = get_pixbuf("close.png")
+        icon = get_icon_pixbuf("close.png")
         if icon:
             close_button = Gtk.Image()
             close_button.set_from_pixbuf(icon)
@@ -272,10 +265,7 @@ class Popup(Gtk.Window):
 
 
     def action_button(self, action_id, action_text):
-        try:
-            button = Gtk.Button(action_text.decode("utf-8"))
-        except Exception:
-            button = Gtk.Button(bytestostr(action_text))
+        button = Gtk.Button(u(action_text))
         button.set_relief(Gtk.ReliefStyle.NORMAL)
         def popup_cb_clicked(*args):
             self.hide_notification()
@@ -308,7 +298,7 @@ class Popup(Gtk.Window):
         """Move the notification window down, when an older notification is removed"""
         log("reposition(%s, %s)", offset, stack)
         new_offset = self.h + offset
-        self.move(self.get_x(self.w), self.get_y(new_offset))
+        GLib.idle_add(self.move, self.get_x(self.w), self.get_y(new_offset))
         return new_offset
 
     def fade_in(self):
@@ -361,6 +351,8 @@ class Popup(Gtk.Window):
             if v:
                 setattr(self, timer, None)
                 GLib.source_remove(v)
+        #destroy window from the UI thread:
+        GLib.idle_add(self.destroy)
         self.destroy()
         self.destroy_cb(self)
 
