@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2011-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -14,7 +14,7 @@ gi.require_version("Pango", "1.0")
 gi.require_version("GdkPixbuf", "2.0")
 from gi.repository import GLib, GdkPixbuf, Pango, GObject, Gtk, Gdk     #@UnresolvedImport
 
-from xpra.util import iround, first_time, envint, envbool
+from xpra.util import first_time, envint, envbool
 from xpra.os_util import strtobytes, WIN32, OSX
 from xpra.log import Logger
 
@@ -25,6 +25,7 @@ alphalog = Logger("gtk", "alpha")
 SHOW_ALL_VISUALS = False
 #try to get workarea from GTK:
 GTK_WORKAREA = envbool("XPRA_GTK_WORKAREA", True)
+SMOOTH_SCROLL = envbool("XPRA_SMOOTH_SCROLL", True)
 
 GTK_VERSION_INFO = {}
 def get_gtk_version_info() -> dict:
@@ -75,9 +76,13 @@ def pixbuf_save_to_memory(pixbuf, fmt="png") -> bytes:
     return b"".join(buf)
 
 
-def GDKWindow(parent=None, width=1, height=1, window_type=Gdk.WindowType.TOPLEVEL,
-              event_mask=0, wclass=Gdk.WindowWindowClass.INPUT_OUTPUT, title=None,
-              x=None, y=None, override_redirect=False, visual=None) -> Gdk.Window:
+def GDKWindow(*args, **kwargs) -> Gdk.Window:
+    return new_GDKWindow(Gdk.Window, *args, **kwargs)
+
+def new_GDKWindow(gdk_window_class,
+                  parent=None, width=1, height=1, window_type=Gdk.WindowType.TOPLEVEL,
+                  event_mask=0, wclass=Gdk.WindowWindowClass.INPUT_OUTPUT, title=None,
+                  x=None, y=None, override_redirect=False, visual=None) -> Gdk.Window:
     attributes_mask = 0
     attributes = Gdk.WindowAttr()
     if x is not None:
@@ -105,7 +110,7 @@ def GDKWindow(parent=None, width=1, height=1, window_type=Gdk.WindowType.TOPLEVE
     #wclass:
     attributes.wclass = wclass
     mask = Gdk.WindowAttributesType(attributes_mask)
-    return Gdk.Window(parent, attributes, mask)
+    return gdk_window_class(parent, attributes, mask)
 
 def set_visual(window, alpha : bool=True) -> bool:
     screen = window.get_screen()
@@ -194,7 +199,10 @@ BUTTON_MASK = {
 em = Gdk.EventMask
 WINDOW_EVENT_MASK = em.STRUCTURE_MASK | em.KEY_PRESS_MASK | em.KEY_RELEASE_MASK \
         | em.POINTER_MOTION_MASK | em.BUTTON_PRESS_MASK | em.BUTTON_RELEASE_MASK \
-        | em.PROPERTY_CHANGE_MASK | em.SCROLL_MASK
+        | em.PROPERTY_CHANGE_MASK | em.SCROLL_MASK | em.SMOOTH_SCROLL_MASK
+if SMOOTH_SCROLL:
+    WINDOW_EVENT_MASK |= em.SMOOTH_SCROLL_MASK
+
 del em
 
 
@@ -237,9 +245,9 @@ def get_screens_info() -> dict:
 def get_screen_sizes(xscale=1, yscale=1):
     from xpra.platform.gui import get_workarea, get_workareas
     def xs(v):
-        return iround(v/xscale)
+        return round(v/xscale)
     def ys(v):
-        return iround(v/yscale)
+        return round(v/yscale)
     def swork(*workarea):
         return xs(workarea[0]), ys(workarea[1]), xs(workarea[2]), ys(workarea[3])
     display = Gdk.Display.get_default()
@@ -250,123 +258,81 @@ def get_screen_sizes(xscale=1, yscale=1):
     def dpi(size_pixels, size_mm):
         if size_mm==0:
             return 0
-        return iround(size_pixels * 254 / size_mm / 10)
-    n_screens = display.get_n_screens()
-    get_n_monitors = getattr(display, "get_n_monitors", None)
-    if get_n_monitors:
-        #GTK 3.22: always just one screen
-        n_monitors = get_n_monitors()
-        workareas = get_workareas()
-        if workareas and len(workareas)!=n_monitors:
-            screenlog(" workareas: %s", workareas)
-            screenlog(" number of monitors does not match number of workareas!")
-            workareas = []
-        monitors = []
-        for j in range(n_monitors):
-            monitor = display.get_monitor(j)
-            geom = monitor.get_geometry()
-            manufacturer, model = monitor.get_manufacturer(), monitor.get_model()
-            if manufacturer and model:
-                plug_name = "%s %s" % (manufacturer, model)
-            elif manufacturer:
-                plug_name = manufacturer
-            elif model:
-                plug_name = model
-            else:
-                plug_name = "%i" % j
-            wmm, hmm = monitor.get_width_mm(), monitor.get_height_mm()
-            monitor_info = [plug_name, xs(geom.x), ys(geom.y), xs(geom.width), ys(geom.height), wmm, hmm]
-            screenlog(" monitor %s: %s", j, monitor)
-            if GTK_WORKAREA and hasattr(monitor, "get_workarea"):
-                rect = monitor.get_workarea()
-                monitor_info += list(swork(rect.x, rect.y, rect.width, rect.height))
-            elif workareas:
-                w = workareas[j]
-                monitor_info += list(swork(*w))
-            monitors.append(tuple(monitor_info))
-        screen = display.get_default_screen()
-        sw, sh = screen.get_width(), screen.get_height()
-        work_x, work_y, work_width, work_height = swork(0, 0, sw, sh)
-        workarea = get_workarea()   #pylint: disable=assignment-from-none
-        if workarea:
-            work_x, work_y, work_width, work_height = swork(*workarea)  #pylint: disable=not-an-iterable
-        screenlog(" workarea=%s", workarea)
-        wmm = screen.get_width_mm()
-        hmm = screen.get_height_mm()
-        xdpi = dpi(sw, wmm)
-        ydpi = dpi(sh, hmm)
+        return round(size_pixels * 254 / size_mm / 10)
+    #GTK 3.22 onwards always returns just a single screen,
+    #potentially with multiple monitors
+    n_monitors = display.get_n_monitors()
+    workareas = get_workareas()
+    if workareas and len(workareas)!=n_monitors:
+        screenlog(" workareas: %s", workareas)
+        screenlog(" number of monitors does not match number of workareas!")
+        workareas = []
+    monitors = []
+    for j in range(n_monitors):
+        monitor = display.get_monitor(j)
+        geom = monitor.get_geometry()
+        manufacturer, model = monitor.get_manufacturer(), monitor.get_model()
+        if manufacturer=="unknown":
+            manufacturer = ""
+        if model=="unknown":
+            model = ""
+        if manufacturer and model:
+            plug_name = "%s %s" % (manufacturer, model)
+        elif manufacturer:
+            plug_name = manufacturer
+        elif model:
+            plug_name = model
+        else:
+            plug_name = "%i" % j
+        wmm, hmm = monitor.get_width_mm(), monitor.get_height_mm()
+        monitor_info = [plug_name, xs(geom.x), ys(geom.y), xs(geom.width), ys(geom.height), wmm, hmm]
+        screenlog(" monitor %s: %s, model=%s, manufacturer=%s",
+                  j, type(monitor).__name__, monitor.get_model(), monitor.get_manufacturer())
+        if GTK_WORKAREA and hasattr(monitor, "get_workarea"):
+            rect = monitor.get_workarea()
+            monitor_info += list(swork(rect.x, rect.y, rect.width, rect.height))
+        elif workareas:
+            w = workareas[j]
+            monitor_info += list(swork(*w))
+        monitors.append(tuple(monitor_info))
+    screen = display.get_default_screen()
+    sw, sh = screen.get_width(), screen.get_height()
+    work_x, work_y, work_width, work_height = swork(0, 0, sw, sh)
+    workarea = get_workarea()   #pylint: disable=assignment-from-none
+    if workarea:
+        work_x, work_y, work_width, work_height = swork(*workarea)  #pylint: disable=not-an-iterable
+    screenlog(" workarea=%s", workarea)
+    wmm = screen.get_width_mm()
+    hmm = screen.get_height_mm()
+    xdpi = dpi(sw, wmm)
+    ydpi = dpi(sh, hmm)
+    if xdpi<MIN_DPI or xdpi>MAX_DPI or ydpi<MIN_DPI or ydpi>MAX_DPI:
+        log("ignoring invalid screen size %ix%imm", wmm, hmm)
+        if os.environ.get("WAYLAND_DISPLAY"):
+            log(" (wayland display?)")
+        if n_monitors>0:
+            wmm = 0
+            for mi in range(n_monitors):
+                monitor = display.get_monitor(mi)
+                log(" monitor %i: %s, model=%s, manufacturer=%s",
+                    mi, monitor, monitor.get_model(), monitor.get_manufacturer())
+                wmm += monitor.get_width_mm()
+                hmm += monitor.get_height_mm()
+            wmm /= n_monitors
+            hmm /= n_monitors
+            xdpi = dpi(sw, wmm)
+            ydpi = dpi(sh, hmm)
         if xdpi<MIN_DPI or xdpi>MAX_DPI or ydpi<MIN_DPI or ydpi>MAX_DPI:
-            log("ignoring invalid screen size %ix%imm", wmm, hmm)
-            if os.environ.get("WAYLAND_DISPLAY"):
-                log(" (wayland display?)")
-            if n_monitors>0:
-                wmm = sum(display.get_monitor(i).get_width_mm() for i in range(n_monitors))
-                hmm = sum(display.get_monitor(i).get_height_mm() for i in range(n_monitors))
-                xdpi = dpi(sw, wmm)
-                ydpi = dpi(sh, hmm)
-            if xdpi<MIN_DPI or xdpi>MAX_DPI or ydpi<MIN_DPI or ydpi>MAX_DPI:
-                #still invalid, generate one from DPI=96
-                wmm = iround(sw*25.4/96)
-                hmm = iround(sh*25.4/96)
-            log(" using %ix%i mm", wmm, hmm)
-        item = (screen.make_display_name(), xs(sw), ys(sh),
-                    wmm, hmm,
-                    monitors,
-                    work_x, work_y, work_width, work_height)
-        screenlog(" screen: %s", item)
-        screen_sizes = [item]
-    else:
-        i=0
-        screen_sizes = []
-        #GTK2 or GTK3<3.22:
-        screenlog("get_screen_sizes(%f, %f) found %s screens", xscale, yscale, n_screens)
-        while i<n_screens:
-            screen = display.get_screen(i)
-            j = 0
-            monitors = []
-            workareas = []
-            #native "get_workareas()" is only valid for a single screen (but describes all the monitors)
-            #and it is only implemented on win32 right now
-            #other platforms only implement "get_workarea()" instead, which is reported against the screen
-            n_monitors = screen.get_n_monitors()
-            screenlog(" screen %s has %s monitors", i, n_monitors)
-            if n_screens==1:
-                workareas = get_workareas()
-                if workareas and len(workareas)!=n_monitors:
-                    screenlog(" workareas: %s", workareas)
-                    screenlog(" number of monitors does not match number of workareas!")
-                    workareas = []
-            while j<screen.get_n_monitors():
-                geom = screen.get_monitor_geometry(j)
-                plug_name = ""
-                if hasattr(screen, "get_monitor_plug_name"):
-                    plug_name = screen.get_monitor_plug_name(j) or ""
-                wmm = -1
-                if hasattr(screen, "get_monitor_width_mm"):
-                    wmm = screen.get_monitor_width_mm(j)
-                hmm = -1
-                if hasattr(screen, "get_monitor_height_mm"):
-                    hmm = screen.get_monitor_height_mm(j)
-                monitor = [plug_name, xs(geom.x), ys(geom.y), xs(geom.width), ys(geom.height), wmm, hmm]
-                screenlog(" monitor %s: %s", j, monitor)
-                if workareas:
-                    w = workareas[j]
-                    monitor += list(swork(*w))
-                monitors.append(tuple(monitor))
-                j += 1
-            work_x, work_y, work_width, work_height = swork(0, 0, screen.get_width(), screen.get_height())
-            workarea = get_workarea()   #pylint: disable=assignment-from-none
-            if workarea:
-                work_x, work_y, work_width, work_height = swork(*workarea)  #pylint: disable=not-an-iterable
-            screenlog(" workarea=%s", workarea)
-            item = (screen.make_display_name(), xs(screen.get_width()), ys(screen.get_height()),
-                        screen.get_width_mm(), screen.get_height_mm(),
-                        monitors,
-                        work_x, work_y, work_width, work_height)
-            screenlog(" screen %s: %s", i, item)
-            screen_sizes.append(item)
-            i += 1
-    return screen_sizes
+            #still invalid, generate one from DPI=96
+            wmm = round(sw*25.4/96)
+            hmm = round(sh*25.4/96)
+        log(" using %ix%i mm", wmm, hmm)
+    screen0 = (screen.make_display_name(), xs(sw), ys(sh),
+                wmm, hmm,
+                monitors,
+                work_x, work_y, work_width, work_height)
+    screenlog(" screen: %s", screen0)
+    return [screen0]
 
 def get_screen_info(display, screen) -> dict:
     info = {}
@@ -581,7 +547,22 @@ def get_icon_from_file(filename):
     return pixbuf
 
 
-def imagebutton(title, icon, tooltip=None, clicked_callback=None, icon_size=32,
+def get_icon_pixbuf(icon_name):
+    try:
+        if not icon_name:
+            log("get_icon_pixbuf(%s)=None", icon_name)
+            return None
+        from xpra.platform.paths import get_icon_filename
+        icon_filename = get_icon_filename(icon_name)
+        log("get_pixbuf(%s) icon_filename=%s", icon_name, icon_filename)
+        if icon_filename:
+            return GdkPixbuf.Pixbuf.new_from_file(icon_filename)
+    except Exception:
+        log.error("get_icon_pixbuf(%s)", icon_name, exc_info=True)
+    return None
+
+
+def imagebutton(title, icon=None, tooltip=None, clicked_callback=None, icon_size=32,
                 default=False, min_size=None, label_color=None, label_font=None) -> Gtk.Button:
     button = Gtk.Button(title)
     settings = button.get_settings()
@@ -599,17 +580,17 @@ def imagebutton(title, icon, tooltip=None, clicked_callback=None, icon_size=32,
     if default:
         button.set_can_default(True)
     if label_color or label_font:
+        l = button
         try:
             alignment = button.get_children()[0]
             b_hbox = alignment.get_children()[0]
             l = b_hbox.get_children()[1]
-        except IndexError:
+        except (IndexError, AttributeError):
             pass
-        else:
-            if label_color:
-                l.modify_fg(Gtk.StateType.NORMAL, label_color)
-            if label_font:
-                l.modify_font(label_font)
+        if label_color and hasattr(l, "modify_fg"):
+            l.modify_fg(Gtk.StateType.NORMAL, label_color)
+        if label_font and hasattr(l, "modify_font"):
+            l.modify_font(label_font)
     return button
 
 def menuitem(title, image=None, tooltip=None, cb=None) -> Gtk.ImageMenuItem:
@@ -631,16 +612,6 @@ def menuitem(title, image=None, tooltip=None, cb=None) -> Gtk.ImageMenuItem:
     return menu_item
 
 
-def get_icon_pixbuf(icon_name):
-    if not icon_name:
-        return None
-    from xpra.platform.paths import get_icon_filename
-    icon_filename = get_icon_filename(icon_name)
-    if icon_filename and os.path.exists(icon_filename):
-        return GdkPixbuf.Pixbuf.new_from_file(icon_filename)
-    return None
-
-
 def add_close_accel(window, callback):
     accel_groups = []
     def wa(s, cb):
@@ -697,7 +668,7 @@ class TableBuilder:
                 i += 1
         self.inc()
 
-    def attach(self, widget, i, count=1,
+    def attach(self, widget, i=0, count=1,
                xoptions=Gtk.AttachOptions.FILL, yoptions=Gtk.AttachOptions.FILL,
                xpadding=10, ypadding=0):
         self.table.attach(widget, i, i+count, self.row, self.row+1,
