@@ -36,8 +36,8 @@ from xpra.clipboard.clipboard_core import (
     ClipboardProxyCore, log, _filter_targets,
     TEXT_TARGETS, MAX_CLIPBOARD_PACKET_SIZE,
     )
-from xpra.util import csv, ellipsizer, envint, envbool, roundup
-from xpra.os_util import bytestostr, strtobytes
+from xpra.util import net_utf8, csv, ellipsizer, envint, envbool, roundup
+from xpra.os_util import bytestostr
 from xpra.platform.win32.constants import PROCESS_QUERY_INFORMATION
 
 CP_UTF8 = 65001
@@ -116,10 +116,12 @@ log("win32 clipboard: RETRY=%i, DELAY=%i, CONVERT_LINE_ENDINGS=%s",
     RETRY, DELAY, CONVERT_LINE_ENDINGS)
 #can be used to blacklist problematic clipboard peers:
 #ie: VBoxTray.exe
-BLACKLISTED_CLIPBOARD_CLIENTS = [x for x in
-                                 os.environ.get("XPRA_BLACKLISTED_CLIPBOARD_CLIENTS", "").split(",")
-                                 if x]
+def get_clients(envvar, default=""):
+    return tuple(x for x in os.environ.get("XPRA_%s_CLIPBOARD_CLIENTS" % envvar, default).split(",") if x)
+BLACKLISTED_CLIPBOARD_CLIENTS = get_clients("BLACKLISTED")
+SYNCDELAY_CLIPBOARD_CLIENTS = get_clients("NOSYNC", "VBoxTray.exe")
 log("BLACKLISTED_CLIPBOARD_CLIENTS=%s", BLACKLISTED_CLIPBOARD_CLIENTS)
+log("SYNCDELAY_CLIPBOARD_CLIENTS=%s", SYNCDELAY_CLIPBOARD_CLIENTS)
 COMPRESSED_IMAGES = envbool("XPRA_CLIPBOARD_COMPRESSED_IMAGES", True)
 
 CLIPBOARD_WINDOW_CLASS_NAME = "XpraWin32Clipboard"
@@ -128,6 +130,10 @@ CLIPBOARD_WINDOW_CLASS_NAME = "XpraWin32Clipboard"
 def is_blacklisted(owner_info):
     return any(owner_info.find(x)>=0 for x in BLACKLISTED_CLIPBOARD_CLIENTS)
 
+def is_syncdelay(owner_info):
+    return any(owner_info.find(x)>=0 for x in SYNCDELAY_CLIPBOARD_CLIENTS)
+
+
 #initialize the window we will use
 #for communicating with the OS clipboard API:
 
@@ -217,8 +223,8 @@ def w_to_utf8(data):
         s = buf.raw[:l]
     log("got %i bytes of UNICODE data: %s", len(s), ellipsizer(s))
     if CONVERT_LINE_ENDINGS:
-        return s.decode("utf8").replace("\r\n", "\n").encode("utf8")
-    return strtobytes(s)
+        return s.decode("utf8").replace("\r\n", "\n")
+    return str(s)
 
 
 class Win32Clipboard(ClipboardTimeoutHelper):
@@ -265,14 +271,16 @@ class Win32Clipboard(ClipboardTimeoutHelper):
                 #ie: don't try to sync from VirtualBox
                 log("CLIPBOARDUPDATE coming from '%s' ignored", owner_info)
                 return r
+            min_delay = 500*int(is_syncdelay(owner_info))
+            log("CLIPBOARDUPDATE coming from '%s', min_delay=%i", owner_info, min_delay)
             for proxy in self._clipboard_proxies.values():
                 if not proxy._block_owner_change:
-                    proxy.schedule_emit_token()
+                    proxy.schedule_emit_token(min_delay)
         return r
 
 
     def cleanup(self):
-        ClipboardTimeoutHelper.cleanup(self)
+        super().cleanup()
         self.cleanup_window()
 
     def cleanup_window(self):
@@ -491,7 +499,7 @@ class Win32ClipboardProxy(ClipboardProxyCore):
             return
         self._got_token_events += 1
         log("got token, selection=%s, targets=%s, target data=%s, claim=%s, can-receive=%s",
-            self._selection, targets, target_data, claim, self._can_receive)
+            self._selection, targets, ellipsizer(target_data), claim, self._can_receive)
         if self._can_receive:
             self.targets = _filter_targets(targets or ())
             self.target_data = target_data or {}
@@ -525,8 +533,8 @@ class Win32ClipboardProxy(ClipboardProxyCore):
                 #request it:
                 self.send_clipboard_request_handler(self, self._selection, image_formats[0])
         elif dformat==8 and dtype in TEXT_TARGETS:
-            log("we got a byte string: %s", data)
-            self.set_clipboard_text(data)
+            log("we got a byte string: %s", ellipsizer(data))
+            self.set_clipboard_text(net_utf8(data))
         elif dformat==8 and dtype.startswith("image/"):
             img_format = dtype.split("/")[-1]   #ie: 'png'
             self.set_clipboard_image(img_format, data)
@@ -676,7 +684,7 @@ class Win32ClipboardProxy(ClipboardProxyCore):
         #convert to wide char
         #get the length in wide chars:
         if CONVERT_LINE_ENDINGS:
-            text = text.decode("utf8").replace("\n", "\r\n").encode("utf8")
+            text = text.replace("\n", "\r\n").encode("utf8")
         wlen = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, text, len(text), None, 0)
         if not wlen:
             self.set_err("failed to prepare to convert to wide char")
@@ -704,10 +712,15 @@ class Win32ClipboardProxy(ClipboardProxyCore):
         finally:
             GlobalUnlock(locked)
         #we're going to alter the clipboard ourselves,
-        #ignore messages until we're done:
-        self._block_owner_change = True
+        #ignore messages until we're done,
+        #this may take a while so ensure we do unblock eventually:
+        long_timeout = GLib.timeout_add(2000, self.remove_block)
+        self._block_owner_change = long_timeout
         def cleanup():
-            GLib.idle_add(self.remove_block)
+            if self._block_owner_change==long_timeout:
+                #now safe to re-schedule and run now in the UI thread:
+                self.cancel_unblock()
+                self._block_owner_change = GLib.idle_add(self.remove_block)
         def set_clipboard_data():
             r = EmptyClipboard()
             log("EmptyClipboard()=%s", r)
