@@ -1,17 +1,18 @@
 #!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import sys
 import os
 
-from xpra.util import csv, sorted_nicely
+from xpra.util import csv, sorted_nicely, remove_dupes
 from xpra.os_util import (
     WIN32, OSX, POSIX,
     osexpand, getuid, getgid, get_username_for_uid,
-    is_Debian, is_Ubuntu, is_arm,
+    is_Debian, is_Ubuntu, is_CentOS, is_RedHat, is_arm,
+    get_distribution_version_id,
     which,
     )
 
@@ -88,12 +89,12 @@ def get_Xdummy_confdir():
     from xpra.platform.xposix.paths import get_runtime_dir
     xrd = get_runtime_dir()
     if xrd:
-        base = "${XDG_RUNTIME_DIR}/xpra"
+        base = "${XPRA_SESSION_DIR}"
     else:   # pragma: no cover
         base = "${HOME}/.xpra"
     return base+"/xorg.conf.d/$PID"
 
-def get_Xdummy_command(xorg_cmd="Xorg", log_dir="${XPRA_LOG_DIR}", xorg_conf="/etc/xpra/xorg.conf"):
+def get_Xdummy_command(xorg_cmd="Xorg", log_dir="${XPRA_SESSION_DIR}", xorg_conf="${XORG_CONFIG_PREFIX}/etc/xpra/xorg.conf"):
     cmd = [xorg_cmd]    #ie: ["Xorg"] or ["xpra_Xdummy"] or ["./install/bin/xpra_Xdummy"]
     cmd += [
           "-noreset", "-novtswitch",
@@ -102,11 +103,11 @@ def get_Xdummy_command(xorg_cmd="Xorg", log_dir="${XPRA_LOG_DIR}", xorg_conf="/e
           "+extension", "RANDR",
           "+extension", "RENDER",
           "-auth", "$XAUTHORITY",
-          "-logfile", "%s/Xorg.${DISPLAY}.log" % log_dir,
+          "-logfile", "%s/Xorg.log" % log_dir,
           #must be specified with some Xorg versions (ie: arch linux)
           #this directory can store xorg config files, it does not need to be created:
           "-configdir", '"%s"' % get_Xdummy_confdir(),
-          "-config", xorg_conf
+          "-config", '"%s"' % xorg_conf,
           ]
     return cmd
 
@@ -142,6 +143,10 @@ def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir=None,
         #These distros do weird things and this can cause the real X11 server to crash
         #see ticket #2834
         return get_Xvfb_command()
+    if is_CentOS() or is_RedHat():
+        #it's too difficult to build our custom Xdummy on CentOS 9
+        if get_distribution_version_id()=="9":
+            return get_Xvfb_command()
 
     xorg_bin = get_xorg_bin()
     def Xorg_suid_check():
@@ -165,7 +170,7 @@ def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir=None,
                 use_wrapper = True
             else:
                 use_wrapper = False
-        xorg_conf = os.path.join(conf_dir, "xorg.conf")
+        xorg_conf = "${XORG_CONFIG_PREFIX}"+os.path.join(conf_dir, "xorg.conf")
         if use_wrapper:
             xorg_cmd = "xpra_Xdummy"
         else:
@@ -306,14 +311,14 @@ def read_config(conf_file):
         the value for this key will be an array of strings.
     """
     d = {}
-    if not os.path.isfile(conf_file):
-        debug("read_config(%s) is not a file!", conf_file)
+    if not os.path.exists(conf_file) or not os.path.isfile(conf_file):
+        debug("read_config(%s) is not a file or does not exist", conf_file)
         return d
     with open(conf_file, "r") as f:
         lines = []
         no = 0
         for line in f:
-            sline = line.strip().rstrip('\r\n').strip()
+            sline = line.strip().strip('\r\n').strip()
             no += 1
             if not sline:
                 debug("%4s empty line", no)
@@ -471,7 +476,7 @@ def may_create_user_config(xpra_conf_filename=DEFAULT_XPRA_CONF_FILENAME):
                         f.write("\n")
                         f.write("# For more information on the file format,\n")
                         f.write("# see the xpra manual at:\n")
-                        f.write("# https://xpra.org/manual.html\n")
+                        f.write("# https://github.com/Xpra-org/xpra/tree/master/docs/\n")
                         f.write("\n")
                         f.flush()
                     debug("created default config in "+d)
@@ -520,6 +525,7 @@ OPTION_TYPES = {
                     "chdir"             : str,
                     "xvfb"              : str,
                     "socket-dir"        : str,
+                    "sessions-dir"      : str,
                     "mmap"              : str,
                     "log-dir"           : str,
                     "log-file"          : str,
@@ -600,6 +606,7 @@ OPTION_TYPES = {
                     "use-display"       : str,
                     "fake-xinerama"     : str,
                     "resize-display"    : str,
+                    "reconnect"         : bool,
                     "tray"              : bool,
                     "pulseaudio"        : bool,
                     "dbus-proxy"        : bool,
@@ -650,7 +657,9 @@ OPTION_TYPES = {
                     "source"            : list,
                     "source-start"      : list,
                     "start"             : list,
+                    "start-late"        : list,
                     "start-child"       : list,
+                    "start-child-late"  : list,
                     "start-after-connect"       : list,
                     "start-child-after-connect" : list,
                     "start-on-connect"          : list,
@@ -660,7 +669,6 @@ OPTION_TYPES = {
                     "bind"              : list,
                     "bind-vsock"        : list,
                     "bind-tcp"          : list,
-                    "bind-udp"          : list,
                     "bind-ws"           : list,
                     "bind-wss"          : list,
                     "bind-ssl"          : list,
@@ -669,7 +677,6 @@ OPTION_TYPES = {
                     "auth"              : list,
                     "vsock-auth"        : list,
                     "tcp-auth"          : list,
-                    "udp-auth"          : list,
                     "ws-auth"           : list,
                     "wss-auth"          : list,
                     "ssl-auth"          : list,
@@ -695,17 +702,19 @@ NON_COMMAND_LINE_OPTIONS = [
 
 START_COMMAND_OPTIONS = [
     "start", "start-child",
+    "start-late", "start-child-late",
     "start-after-connect", "start-child-after-connect",
     "start-on-connect", "start-child-on-connect",
     "start-on-last-client-exit", "start-child-on-last-client-exit",
     ]
-BIND_OPTIONS = ["bind", "bind-tcp", "bind-udp", "bind-ssl", "bind-ws", "bind-wss", "bind-vsock", "bind-rfb"]
+BIND_OPTIONS = ["bind", "bind-tcp", "bind-ssl", "bind-ws", "bind-wss", "bind-vsock", "bind-rfb"]
 
 #keep track of the options added since v3,
 #so we can generate command lines that work with older supported versions:
 OPTIONS_ADDED_SINCE_V3 = [
     "source", "source-start", "headerbar",
-    "splash", "http-scripts", "client-socket-dirs",
+    "splash", "sessions-dir", "http-scripts", "reconnect", "client-socket-dirs",
+    "start-late", "start-child-late",
     ]
 OPTIONS_COMPAT_NAMES = {
     "--compression_level=" : "-z"
@@ -749,6 +758,7 @@ CLIENT_OPTIONS = ["title", "username", "password", "session-name",
 CLIENT_ONLY_OPTIONS = ["username", "swap-keys", "dock-icon",
                        "tray", "delay-tray", "tray-icon",
                        "attach",
+                       "reconnect",
                        ]
 
 #options that clients can pass to the proxy
@@ -778,13 +788,16 @@ PROXY_START_OVERRIDABLE_OPTIONS = [
     "splash",
     "printing", "file-transfer", "open-command", "open-files", "open-url", "start-new-commands",
     "mmap", "mmap-group", "mdns",
-    "auth", "vsock-auth", "tcp-auth", "udp-auth", "ws-auth", "wss-auth", "ssl-auth", "ssh-auth", "rfb-auth",
-    "bind", "bind-vsock", "bind-tcp", "bind-udp", "bind-ssl", "bind-ws", "bind-wss", "bind-ssh", "bind-rfb",
+    "auth", "vsock-auth", "tcp-auth", "ws-auth", "wss-auth", "ssl-auth", "ssh-auth", "rfb-auth",
+    "bind", "bind-vsock", "bind-tcp", "bind-ssl", "bind-ws", "bind-wss", "bind-ssh", "bind-rfb",
     "rfb-upgrade", "bandwidth-limit",
-    "start", "start-child", "source", "source-start",
+    "start", "start-child",
+    "start-late", "start-child-late",
+    "source", "source-start",
     "start-after-connect", "start-child-after-connect",
     "start-on-connect", "start-child-on-connect",
     "start-on-last-client-exit", "start-child-on-last-client-exit",
+    "sessions-dir",
     ]
 tmp = os.environ.get("XPRA_PROXY_START_OVERRIDABLE_OPTIONS", "")
 if tmp:
@@ -861,9 +874,12 @@ def get_defaults():
         return GLOBAL_DEFAULTS
     from xpra.platform.features import (
         OPEN_COMMAND, DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS,
-        DEFAULT_ENV, CAN_DAEMONIZE, SYSTEM_PROXY_SOCKET,
+        SOURCE, DEFAULT_ENV, CAN_DAEMONIZE, SYSTEM_PROXY_SOCKET,
+        )
+    from xpra.platform.paths import (
+        get_download_dir, get_remote_run_xpra_scripts,
+        get_sessions_dir, get_socket_dirs, get_client_socket_dirs,
         )
-    from xpra.platform.paths import get_download_dir, get_remote_run_xpra_scripts, get_socket_dirs, get_client_socket_dirs
     try:
         from xpra.platform.info import get_username
         username = get_username()
@@ -873,9 +889,9 @@ def get_defaults():
     build_root = os.environ.get("RPM_BUILD_ROOT")
     if build_root:
         conf_dirs.append(os.path.join(build_root, "etc", "xpra"))
-    xpra_cmd = sys.argv[0]
     bin_dir = None
     if sys.argv:
+        xpra_cmd = sys.argv[0]
         for strip in ("/usr/bin", "/bin"):
             pos = xpra_cmd.find(strip)
             if pos>=0:
@@ -899,7 +915,7 @@ def get_defaults():
         #try to find it quickly:
         try:
             import glob
-            fake_xinerama = sorted_nicely(glob.glob("/usr/lib*/libfakeXinerama.so.*"))[-1]
+            fake_xinerama = sorted_nicely(glob.glob("/lib*/libfakeXinerama.so.*"))[-1]
         except Exception:
             fake_xinerama = "auto"
     else:
@@ -932,7 +948,7 @@ def get_defaults():
                     "tcp-encryption"    : "",
                     "encryption-keyfile": "",
                     "tcp-encryption-keyfile": "",
-                    "pidfile"           : "",
+                    "pidfile"           : "${XPRA_SESSION_DIR}/server.pid",
                     "ssh"               : "auto",
                     "systemd-run"       : get_default_systemd_run(),
                     "systemd-run-args"  : "",
@@ -940,8 +956,9 @@ def get_defaults():
                     "xvfb"              : xvfb_str,
                     "chdir"             : "",
                     "socket-dir"        : "",
+                    "sessions-dir"      : get_sessions_dir(),
                     "log-dir"           : "auto",
-                    "log-file"          : "$DISPLAY.log",
+                    "log-file"          : "server.log",
                     "border"            : "auto,5:off",
                     "window-close"      : "auto",
                     "min-size"          : "",
@@ -958,7 +975,7 @@ def get_defaults():
                     "pdf-printer"       : "",
                     "postscript-printer": DEFAULT_POSTSCRIPT_PRINTER,
                     "debug"             : "",
-                    "input-method"      : "none",
+                    "input-method"      : "auto",
                     "sound-source"      : "",
                     "html"              : "auto",
                     "http-scripts"      : "all",
@@ -990,9 +1007,9 @@ def get_defaults():
                     "ssl-server-hostname": "",
                     "ssl-options"       : "ALL,NO_COMPRESSION",
                     "quality"           : 0,
-                    "min-quality"       : 30,
+                    "min-quality"       : 1,
                     "speed"             : 0,
-                    "min-speed"         : 30,
+                    "min-speed"         : 1,
                     "compression_level" : 1,
                     "dpi"               : 0,
                     "file-size-limit"   : "100M",
@@ -1011,6 +1028,7 @@ def get_defaults():
                     "use-display"       : "auto",
                     "fake-xinerama"     : fake_xinerama,
                     "resize-display"    : ["no", "yes"][not OSX and not WIN32],
+                    "reconnect"         : True,
                     "tray"              : True,
                     "pulseaudio"        : DEFAULT_PULSEAUDIO,
                     "dbus-proxy"        : not OSX and not WIN32,
@@ -1055,7 +1073,7 @@ def get_defaults():
                     "client-socket-dirs" : get_client_socket_dirs(),
                     "remote-xpra"       : get_remote_run_xpra_scripts(),
                     "encodings"         : ["all"],
-                    "proxy-video-encoders" : [],
+                    "proxy-video-encoders" : ["none"],
                     "video-encoders"    : ["all"],
                     "csc-modules"       : ["all"],
                     "video-decoders"    : ["all"],
@@ -1067,7 +1085,6 @@ def get_defaults():
                     "bind"              : ["auto"],
                     "bind-vsock"        : [],
                     "bind-tcp"          : [],
-                    "bind-udp"          : [],
                     "bind-ws"           : [],
                     "bind-wss"          : [],
                     "bind-ssl"          : [],
@@ -1076,17 +1093,18 @@ def get_defaults():
                     "auth"              : [],
                     "vsock-auth"        : [],
                     "tcp-auth"          : [],
-                    "udp-auth"          : [],
                     "ws-auth"           : [],
                     "wss-auth"          : [],
                     "ssl-auth"          : [],
                     "ssh-auth"          : [],
                     "rfb-auth"          : [],
                     "password-file"     : [],
-                    "source"            : [],
+                    "source"            : SOURCE,
                     "source-start"      : [],
                     "start"             : [],
+                    "start-late"        : [],
                     "start-child"       : [],
+                    "start-child-late"  : [],
                     "start-after-connect"       : [],
                     "start-child-after-connect" : [],
                     "start-on-connect"          : [],
@@ -1313,10 +1331,9 @@ def _csvstr(value):
     raise Exception("don't know how to convert %s to a csv list!" % type(value))
 
 def _nodupes(s):
-    from xpra.util import remove_dupes
     return remove_dupes(x.strip().lower() for x in s.split(","))
 
-def fixup_video_all_or_none(options):
+def fixup_video_all_or_none(options, defaults=None):
     #we import below, but only if we really need to access
     #one of the lists, because those are expensive to probe
     #(we have to load the codec, which may load other libraries)
@@ -1328,21 +1345,30 @@ def fixup_video_all_or_none(options):
     # HARDWARE_ENCODER_OPTIONS,
     #)
     def getlist(strarg, help_txt, all_list_name):
-        if strarg=="help":
-            from xpra.codecs import video_helper
+        if strarg=="none":
+            v = []
+        elif strarg=="all":
+            from xpra.codecs import video_helper    #pylint: disable=import-outside-toplevel
+            v = getattr(video_helper, all_list_name)
+        else:
+            v = [x for x in _nodupes(strarg) if x]
+        if "help" in v:
+            from xpra.codecs import video_helper    #@Reimport pylint: disable=import-outside-toplevel
             raise InitInfo("the following %s may be available: %s" %
                            (help_txt, csv(getattr(video_helper, all_list_name))))
-        elif strarg=="none":
-            return []
-        elif strarg=="all":
-            from xpra.codecs import video_helper    #@Reimport
-            return getattr(video_helper, all_list_name)
-        else:
-            return [x for x in _nodupes(strarg) if x]
-    vestr   = _csvstr(options.video_encoders)
-    cscstr  = _csvstr(options.csc_modules)
-    vdstr   = _csvstr(options.video_decoders)
-    pvestr  = _csvstr(options.proxy_video_encoders)
+        return v
+    vestr = _csvstr(options.video_encoders)
+    if not vestr and defaults:
+        vestr = _csvstr(defaults.video_encoders)
+    cscstr = _csvstr(options.csc_modules)
+    if not cscstr and defaults:
+        cscstr = _csvstr(defaults.csc_modules)
+    vdstr = _csvstr(options.video_decoders)
+    if not vdstr and defaults:
+        vdstr = _csvstr(defaults.video_decoders)
+    pvestr = _csvstr(options.proxy_video_encoders)
+    if not pvestr and defaults:
+        pvestr  = _csvstr(defaults.proxy_video_encoders)
     options.video_encoders  = getlist(vestr,    "video encoders",   "ALL_VIDEO_ENCODER_OPTIONS")
     options.csc_modules     = getlist(cscstr,   "csc modules",      "ALL_CSC_MODULE_OPTIONS")
     options.video_decoders  = getlist(vdstr,    "video decoders",   "ALL_VIDEO_DECODER_OPTIONS")
@@ -1374,24 +1400,45 @@ def fixup_pings(options):
         options.pings = 5
 
 def fixup_encodings(options):
-    try:
-        from xpra.codecs.codec_constants import PREFERRED_ENCODING_ORDER
-    except ImportError:
-        return
-    RENAME = {"jpg" : "jpeg"}
-    if options.encoding:
-        options.encoding = RENAME.get(options.encoding, options.encoding)
+    from xpra.codecs.codec_constants import PREFERRED_ENCODING_ORDER
     estr = _csvstr(options.encodings)
-    if estr=="all":
-        #replace with an actual list
-        options.encodings = list(PREFERRED_ENCODING_ORDER)
-        return
+    RENAME = {"jpg" : "jpeg"}
+    options.encoding = RENAME.get(options.encoding, options.encoding)
     encodings = [RENAME.get(x, x) for x in _nodupes(estr)]
+    while True:
+        try:
+            i = encodings.index("all")
+        except ValueError:
+            break
+        else:
+            #replace 'all' with the actual value:
+            encodings = encodings[:i]+list(PREFERRED_ENCODING_ORDER)+encodings[i+1:]
+    #if the list only has items to exclude (ie: '-scroll,-jpeg')
+    #then 'all' is implied:
+    if not any(True for e in encodings if not e.startswith("-")):
+        encodings = list(PREFERRED_ENCODING_ORDER)+encodings
     if "rgb" in encodings:
         if "rgb24" not in encodings:
             encodings.append("rgb24")
         if "rgb32" not in encodings:
             encodings.append("rgb32")
+    encodings = remove_dupes(encodings)
+    invalid = tuple(e.lstrip("-") for e in encodings if (e.lstrip("-") not in PREFERRED_ENCODING_ORDER))
+    if invalid:
+        from xpra.exit_codes import EXIT_UNSUPPORTED  #pylint: disable=import-outside-toplevel
+        raise InitExit(EXIT_UNSUPPORTED, "invalid encodings specified: %s" % csv(invalid))
+    #now we have a list of encodings, but some of them may be prefixed with "-"
+    for rm in tuple(e for e in encodings if e.startswith("-")):
+        while True:
+            try:
+                encodings.remove(rm)
+            except ValueError:
+                break
+        while True:
+            try:
+                encodings.remove(rm[1:])
+            except ValueError:
+                break
     options.encodings = encodings
 
 def fixup_compression(options):
@@ -1426,7 +1473,6 @@ def fixup_keyboard(options):
     #variants and layouts can be specified as CSV, convert them to lists:
     def p(v):
         try:
-            from xpra.util import remove_dupes
             r = remove_dupes(x.strip() for x in v.split(","))
             #remove empty string if that's the only value:
             if r and len(r)==1 and r[0]=="":
@@ -1483,13 +1529,24 @@ def abs_paths(options):
             setattr(options, f, os.path.abspath(v))
 
 
-def fixup_options(options, skip_encodings=False):
+def fixup_options(options, defaults=None, skip_encodings=False):
     if not skip_encodings:
         fixup_encodings(options)
+        fixup_video_all_or_none(options, defaults)
+    else:
+        for setting, default_value in {
+            "video_encoders"        : ["all"],
+            "video_decoders"        : ["all"],
+            "csc_modules"           : ["all"],
+            "proxy_video_encoders"  : ["none"],
+            "encodings"             : ["all"],
+            }.items():
+            v = getattr(options, setting)
+            if not v:
+                setattr(options, setting, default_value)
     fixup_pings(options)
     fixup_compression(options)
     fixup_packetencoding(options)
-    fixup_video_all_or_none(options)
     fixup_socketdirs(options)
     fixup_clipboard(options)
     fixup_keyboard(options)
