@@ -1,14 +1,14 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 #pylint: disable-msg=E1101
 
 import os.path
 
-from xpra.util import parse_scaling_value, csv, from0to100
-from xpra.os_util import load_binary_file, bytestostr, strtobytes
+from xpra.util import parse_scaling_value, csv, from0to100, net_utf8, typedict
+from xpra.os_util import load_binary_file
 from xpra.simple_stats import std_unit
 from xpra.scripts.config import parse_bool, FALSE_OPTIONS, TRUE_OPTIONS
 from xpra.server.control_command import ArgsControlCommand, ControlError
@@ -23,10 +23,10 @@ TOGGLE_FEATURES = (
     )
 
 
-"""
-Control commands for ServerBase
-"""
 class ServerBaseControlCommands(StubServerMixin):
+    """
+    Control commands for ServerBase
+    """
 
     def setup(self):
         self.add_control_commands()
@@ -91,6 +91,8 @@ class ServerBaseControlCommands(StubServerMixin):
             ArgsControlCommand("sound-output",          "control sound forwarding",         min_args=1, max_args=2),
             #windows:
             ArgsControlCommand("workspace",             "move a window to a different workspace", min_args=2, max_args=2, validation=[int, int]),
+            ArgsControlCommand("close",                 "close a window",                   min_args=1, max_args=1, validation=[int]),
+            ArgsControlCommand("delete",                "delete a window",                   min_args=1, max_args=1, validation=[int]),
             ArgsControlCommand("move",                  "move a window",                    min_args=3, max_args=3, validation=[int, int, int]),
             ArgsControlCommand("resize",                "resize a window",                  min_args=3, max_args=3, validation=[int, int, int]),
             ArgsControlCommand("moveresize",            "move and resize a window",         min_args=5, max_args=5, validation=[int, int, int, int, int]),
@@ -288,10 +290,7 @@ class ServerBaseControlCommands(StubServerMixin):
         #we always get the values as strings from the command interface,
         #but those may actually be utf8 encoded binary strings,
         #so we may have to do an ugly roundtrip:
-        try:
-            filename = strtobytes(filename).decode("utf8")
-        except Exception:
-            filename = bytestostr(filename)
+        filename = net_utf8(filename)
         openit = str(openit).lower() in ("open", "true", "1")
         return self.do_control_file_command("send file", client_uuids, filename, "file_transfer", (False, openit))
 
@@ -335,9 +334,15 @@ class ServerBaseControlCommands(StubServerMixin):
         #send it to each client:
         for ss in sources:
             #ie: ServerSource.file_transfer (found in FileTransferAttributes)
-            if not getattr(ss, source_flag_name):
-                log.warn("Warning: cannot %s '%s'", command_type, filename)
-                log.warn(" client %s does not support this feature", ss)
+            if not getattr(ss, source_flag_name, False):
+                #skip the warning if the client is not interactive
+                #(for now just check for 'top' client):
+                if ss.client_type=="top":
+                    l = log
+                else:
+                    l = log.warn
+                l("Warning: cannot %s '%s' to %s client", command_type, filename, ss.client_type)
+                l(" client %s does not support this feature", ss.uuid)
             elif file_size>ss.file_size_limit:
                 log.warn("Warning: cannot %s '%s'", command_type, filename)
                 log.warn(" client %s file size limit is %sB (file is %sB)",
@@ -355,7 +360,7 @@ class ServerBaseControlCommands(StubServerMixin):
         return "removed %i window-filters" % l
 
     def control_command_add_window_filter(self, object_name, property_name, operator, value, client_uuids=""):
-        from xpra.server.window import filters
+        from xpra.server.window import filters  #pylint: disable=import-outside-toplevel
         window_filter = filters.get_window_filter(object_name, property_name, operator, value)
         #log("%s%s=%s", filters.get_window_filter, (object_name, property_name, operator, value), window_filter)
         if client_uuids=="*":
@@ -369,8 +374,8 @@ class ServerBaseControlCommands(StubServerMixin):
 
     def control_command_compression(self, compression):
         c = compression.lower()
-        from xpra.net import compression
-        opts = compression.get_enabled_compressors()    #ie: [lz4, lzo, zlib]
+        from xpra.net import compression    #pylint: disable=import-outside-toplevel
+        opts = compression.get_enabled_compressors()    #ie: [lz4, zlib]
         if c not in opts:
             raise ControlError("compressor argument must be one of: %s" % csv(opts))
         for cproto in tuple(self._server_sources.keys()):
@@ -380,8 +385,8 @@ class ServerBaseControlCommands(StubServerMixin):
 
     def control_command_encoder(self, encoder):
         e = encoder.lower()
-        from xpra.net import packet_encoding
-        opts = packet_encoding.get_enabled_encoders()   #ie: [rencode, bencode, yaml]
+        from xpra.net import packet_encoding  #pylint: disable=import-outside-toplevel
+        opts = packet_encoding.get_enabled_encoders()   #ie: [rencode, rencodeplus, bencode, yaml]
         if e not in opts:
             raise ControlError("encoder argument must be one of: %s" % csv(opts))
         for cproto in tuple(self._server_sources.keys()):
@@ -424,7 +429,7 @@ class ServerBaseControlCommands(StubServerMixin):
         self.mdns_update()
         return "session name set to %s" % name
 
-    def _control_windowsources_from_args(self, *args):
+    def _ws_from_args(self, *args):
         #converts the args to valid window ids,
         #then returns all the window sources for those wids
         if len(args)==0 or len(args)==1 and args[0]=="*":
@@ -435,23 +440,24 @@ class ServerBaseControlCommands(StubServerMixin):
             for x in args:
                 try:
                     wid = int(x)
-                except:
+                except ValueError:
                     raise ControlError("invalid window id: %s" % x) from None
                 if wid in self._id_to_window:
                     wids.append(wid)
                 else:
                     log("window id %s does not exist", wid)
-        wss = {}
+        wss = []
         for csource in tuple(self._server_sources.values()):
             for wid in wids:
                 ws = csource.window_sources.get(wid)
                 window = self._id_to_window.get(wid)
                 if window and ws:
-                    wss[ws] = window
+                    wss.append(ws)
         return wss
 
+
     def _set_encoding_property(self, name, value, *wids):
-        for ws in self._control_windowsources_from_args(*wids).keys():
+        for ws in self._ws_from_args(*wids):
             fn = getattr(ws, "set_%s" % name.replace("-", "_"))   #ie: "set_quality"
             fn(value)
         #now also update the defaults:
@@ -470,35 +476,63 @@ class ServerBaseControlCommands(StubServerMixin):
 
     def control_command_auto_refresh(self, auto_refresh, *wids):
         delay = int(float(auto_refresh)*1000.0)      # ie: 0.5 -> 500 (milliseconds)
-        for ws in self._control_windowsources_from_args(*wids).keys():
+        for ws in self._ws_from_args(*wids):
             ws.set_auto_refresh_delay(auto_refresh)
         return "auto-refresh delay set to %sms for windows %s" % (delay, wids)
 
     def control_command_refresh(self, *wids):
-        for ws in self._control_windowsources_from_args(*wids).keys():
+        for ws in self._ws_from_args(*wids):
             ws.full_quality_refresh({})
         return "refreshed windows %s" % str(wids)
 
     def control_command_scaling_control(self, scaling_control, *wids):
-        for ws in tuple(self._control_windowsources_from_args(*wids).keys()):
+        for ws in tuple(self._ws_from_args(*wids)):
             ws.set_scaling_control(scaling_control)
             ws.refresh()
         return "scaling-control set to %s on windows %s" % (scaling_control, wids)
 
     def control_command_scaling(self, scaling, *wids):
-        for ws in tuple(self._control_windowsources_from_args(*wids).keys()):
+        for ws in tuple(self._ws_from_args(*wids)):
             ws.set_scaling(scaling)
             ws.refresh()
         return "scaling set to %s on windows %s" % (str(scaling), wids)
 
     def control_command_encoding(self, encoding, *args):
+        if encoding in ("add", "remove"):
+            cmd = encoding
+            assert len(args)>0
+            encoding = args[0]
+            wids = args[1:]
+            for ws in tuple(self._ws_from_args(*wids)):
+                encodings = list(ws.encodings)
+                core_encodings = list(ws.core_encodings)
+                for l in (encodings, core_encodings):
+                    if cmd=="add" and encoding not in l:
+                        log("adding %s to %s for %s", encoding, l, ws)
+                        l.append(encoding)
+                    elif cmd=="remove" and encoding in l:
+                        log("removing %s from %s for %s", encoding, l, ws)
+                        l.remove(encoding)
+                    else:
+                        continue
+                ws.encodings = tuple(encodings)
+                ws.core_encodings = tuple(core_encodings)
+                ws.do_set_client_properties(typedict())
+                ws.refresh()
+            return "%s %s" % (["removed", "added"][cmd=="add"], encoding)
+
         strict = None       #means no change
-        if len(args)>0 and args[0] in ("strict", "nostrict"):
+        if encoding in ("strict", "nostrict"):
+            strict = encoding=="strict"
+            encoding = args[0]
+            wids = args[1:]
+        elif len(args)>0 and args[0] in ("strict", "nostrict"):
             #remove "strict" marker
             strict = args[0]=="strict"
-            args = args[1:]
-        wids = args
-        for ws in tuple(self._control_windowsources_from_args(*wids).keys()):
+            wids = args[1:]
+        else:
+            wids = args
+        for ws in tuple(self._ws_from_args(*wids)):
             ws.set_new_encoding(encoding, strict)
             ws.refresh()
         return "set encoding to %s%s for windows %s" % (encoding, ["", " (strict)"][int(strict or 0)], wids)
@@ -531,7 +565,7 @@ class ServerBaseControlCommands(StubServerMixin):
         if wid not in self._id_to_window:
             raise ControlError("invalid window %i" % wid)
         video_subregions = []
-        for ws in self._control_windowsources_from_args(wid).keys():
+        for ws in self._ws_from_args(wid):
             vs = getattr(ws, "video_subregion", None)
             if not vs:
                 log.warn("Warning: cannot set video region enabled flag on window %i:", wid)
@@ -568,11 +602,11 @@ class ServerBaseControlCommands(StubServerMixin):
 
 
     def control_command_lock_batch_delay(self, wid, delay):
-        for ws in self._control_windowsources_from_args(wid).keys():
+        for ws in self._ws_from_args(wid):
             ws.lock_batch_delay(delay)
 
     def control_command_unlock_batch_delay(self, wid):
-        for ws in self._control_windowsources_from_args(wid).keys():
+        for ws in self._ws_from_args(wid):
             ws.unlock_batch_delay()
 
     def control_command_set_lock(self, lock):
@@ -605,9 +639,10 @@ class ServerBaseControlCommands(StubServerMixin):
                 keycode = int(keycode_str, 16)
             else:
                 keycode = int(keycode_str)
-            assert keycode>0 and keycode<=255
-        except:
-            raise ControlError("invalid keycode specified: '%s' (must be a number between 1 and 255)" % keycode_str) from None
+        except ValueError:
+            raise ControlError("invalid keycode specified: '%s' (not a number)" % keycode_str) from None
+        if keycode<=0 or keycode>=255:
+            raise ControlError("invalid keycode value: '%s' (must be between 1 and 255)" % keycode_str)
         if press is not True:
             if press in ("1", "press"):
                 press = True
@@ -636,6 +671,20 @@ class ServerBaseControlCommands(StubServerMixin):
         return "window %s moved to workspace %s" % (wid, workspace)
 
 
+    def control_command_close(self, wid):
+        window = self._id_to_window.get(wid)
+        if not window:
+            raise ControlError("window %s does not exist" % wid)
+        window.request_close()
+        return "requested window %s closed" % window
+
+    def control_command_delete(self, wid):
+        window = self._id_to_window.get(wid)
+        if not window:
+            raise ControlError("window %s does not exist" % wid)
+        window.send_delete()
+        return "requested window %s deleted" % window
+
     def control_command_move(self, wid, x, y):
         window = self._id_to_window.get(wid)
         if not window:
