@@ -7,14 +7,15 @@
 # later version. See the file COPYING for details.
 
 import os
+from time import monotonic
 
 from xpra.server.server_core import ServerCore
 from xpra.server.mixins.server_base_controlcommands import ServerBaseControlCommands
 from xpra.server.background_worker import add_work_item
 from xpra.net.common import may_log_packet
-from xpra.os_util import monotonic_time, bytestostr, strtobytes, WIN32
+from xpra.os_util import bytestostr, strtobytes, WIN32
 from xpra.util import (
-    typedict, flatten_dict, updict, merge_dicts, envbool,
+    typedict, flatten_dict, updict, merge_dicts, envbool, csv,
     SERVER_EXIT, SERVER_ERROR, SERVER_SHUTDOWN, DETACH_REQUEST,
     NEW_CLIENT, DONE, SESSION_BUSY,
     )
@@ -74,6 +75,7 @@ ServerBaseClass = type('ServerBaseClass', SERVER_BASES, {})
 
 log = Logger("server")
 netlog = Logger("network")
+authlog = Logger("auth")
 httplog = Logger("http")
 timeoutlog = Logger("timeout")
 screenlog = Logger("screen")
@@ -111,7 +113,6 @@ class ServerBase(ServerBaseClass):
         self.idle_timeout = 0
         #duplicated from Server Source...
         self.client_shutdown = CLIENT_CAN_SHUTDOWN
-        self.http_stream_check_timers = {}
 
         self.init_packet_handlers()
         self.init_aliases()
@@ -142,9 +143,9 @@ class ServerBase(ServerBaseClass):
         from xpra.scripts import config
         config.warn = log.warn
         for c in SERVER_BASES:
-            start = monotonic_time()
+            start = monotonic()
             c.init(self, opts)
-            end = monotonic_time()
+            end = monotonic()
             log("%3ims in %s.init", 1000*(end-start), c)
         self.sharing = opts.sharing
         self.lock = opts.lock
@@ -154,13 +155,13 @@ class ServerBase(ServerBaseClass):
     def setup(self):
         log("starting component init")
         for c in SERVER_BASES:
-            start = monotonic_time()
+            start = monotonic()
             c.setup(self)
-            end = monotonic_time()
+            end = monotonic()
             log("%3ims in %s.setup", 1000*(end-start), c)
 
     def threaded_init(self):
-        super().threaded_init()
+        super().do_threaded_init()
         log("threaded_init() serverbase start")
         for c in SERVER_BASES:
             if c!=ServerCore:
@@ -169,6 +170,7 @@ class ServerBase(ServerBaseClass):
                 except Exception:
                     log.error("Error during threaded setup of %s", c, exc_info=True)
         log("threaded_init() serverbase end")
+        super().call_init_thread_callbacks()
 
 
     def server_is_ready(self):
@@ -179,9 +181,10 @@ class ServerBase(ServerBaseClass):
     def do_cleanup(self):
         self.server_event("exit")
         self.wait_for_threaded_init()
-        self.cancel_http_stream_check_timers()
+        log("do_cleanup() calling on %s", SERVER_BASES)
         for c in SERVER_BASES:
             if c!=ServerCore:
+                log("%s", c.cleanup)
                 c.cleanup(self)
 
 
@@ -197,9 +200,14 @@ class ServerBase(ServerBaseClass):
 
     ######################################################################
     # shutdown / exit commands:
-    def _process_exit_server(self, _proto, _packet):
-        log.info("Exiting in response to client request")
-        self.cleanup_all_protocols(SERVER_EXIT)
+    def _process_exit_server(self, _proto, packet):
+        reason = SERVER_EXIT
+        message = "Exiting in response to client request"
+        if len(packet)>1:
+            reason = bytestostr(packet[1])
+            message += ": "+reason
+        log.info(message)
+        self.cleanup_all_protocols(reason=reason)
         self.timeout_add(500, self.clean_quit, EXITING_CODE)
 
     def _process_shutdown_server(self, _proto, _packet):
@@ -207,7 +215,7 @@ class ServerBase(ServerBaseClass):
             log.warn("Warning: ignoring shutdown request")
             return
         log.info("Shutting down in response to client request")
-        self.cleanup_all_protocols(SERVER_SHUTDOWN)
+        self.cleanup_all_protocols(reason=SERVER_SHUTDOWN)
         self.timeout_add(500, self.clean_quit)
 
 
@@ -225,24 +233,29 @@ class ServerBase(ServerBaseClass):
         disconnected = 0
         existing_sources = set(ss for p,ss in self._server_sources.items() if p!=proto)
         is_existing_client = uuid and any(ss.uuid==uuid for ss in existing_sources)
-        log("handle_sharing%s lock=%s, sharing=%s, existing sources=%s, is existing client=%s",
+        authlog("handle_sharing%s lock=%s, sharing=%s, existing sources=%s, is existing client=%s",
             (proto, ui_client, detach_request, share, uuid),
             self.lock, self.sharing, existing_sources, is_existing_client)
         #if other clients are connected, verify we can steal or share:
         if existing_sources and not is_existing_client:
             if self.sharing is True or (self.sharing is None and share and all(ss.share for ss in existing_sources)):
-                log("handle_sharing: sharing with %s", tuple(existing_sources))
+                authlog("handle_sharing: sharing with %s", tuple(existing_sources))
             elif self.lock is True:
+                authlog("handle_sharing: session is locked")
                 self.disconnect_client(proto, SESSION_BUSY, "this session is locked")
                 return False, 0, 0
             elif self.lock is not False and any(ss.lock for ss in existing_sources):
+                authlog("handle_sharing: another client has locked the session: %s", csv(ss for ss in existing_sources if ss.lock))
                 self.disconnect_client(proto, SESSION_BUSY, "a client has locked this session")
                 return False, 0, 0
         for p,ss in tuple(self._server_sources.items()):
             if detach_request and p!=proto:
+                authlog("handle_sharing: detaching %s", ss)
                 self.disconnect_client(p, DETACH_REQUEST)
                 disconnected += 1
             elif uuid and ss.uuid==uuid and ui_client and ss.ui_client:
+                authlog("uuid %s is the same as %s", uuid, ss)
+                authlog("existing sources: %s", existing_sources)
                 self.disconnect_client(p, NEW_CLIENT, "new connection from the same uuid")
                 disconnected += 1
             elif ui_client and ss.ui_client:
@@ -266,14 +279,20 @@ class ServerBase(ServerBaseClass):
         #don't accept this connection if we're going to exit-with-client:
         accepted = True
         if disconnected>0 and share_count==0 and self.exit_with_client:
-            self.disconnect_client(proto, SERVER_EXIT, "last client has exited")
+            self.disconnect_client(proto, SERVER_SHUTDOWN, "last client has exited")
             accepted = False
         return accepted, share_count, disconnected
 
     def hello_oked(self, proto, packet, c, auth_caps):
+        if self._server_sources.get(proto):
+            log.warn("Warning: received another 'hello' packet")
+            log.warn(" from an existing connection: %s", proto)
+            return
         if ServerCore.hello_oked(self, proto, packet, c, auth_caps):
             #has been handled
             return
+        if not self.sanity_checks(proto, c):
+            return
         if not c.boolget("steal", True) and self._server_sources:
             self.disconnect_client(proto, SESSION_BUSY, "this session is already active")
             return
@@ -281,20 +300,14 @@ class ServerBase(ServerBaseClass):
             self.send_screenshot(proto)
             return
         #added in 2.2:
-        generic_request = c.strget("request")
+        request = c.strget("request")
         def is_req(mode):
-            return generic_request==mode or c.boolget("%s_request" % mode, False)
-        detach_request  = is_req("detach")
-        stop_request    = is_req("stop_request")
-        exit_request    = is_req("exit_request")
-        event_request   = is_req("event_request")
-        print_request   = is_req("print_request")
-        is_request = detach_request or stop_request or exit_request or event_request or print_request
-        if not is_request:
+            return request==mode or c.boolget("%s_request" % mode, False)
+        if not request:
             #"normal" connection, so log welcome message:
             log.info("Handshake complete; enabling connection")
         else:
-            log("handling request %s", generic_request)
+            log("handling request %s", request)
         self.server_event("handshake-complete")
 
         # Things are okay, we accept this connection, and may disconnect previous one(s)
@@ -302,15 +315,16 @@ class ServerBase(ServerBaseClass):
         ui_client = c.boolget("ui_client", True)
         share = c.boolget("share")
         uuid = c.strget("uuid")
+        detach_request = is_req("detach")
         accepted, share_count, disconnected = self.handle_sharing(proto, ui_client, detach_request, share, uuid)
         if not accepted:
             return
 
-        if detach_request:
+        if is_req("detach"):
             self.disconnect_client(proto, DONE, "%i other clients have been disconnected" % disconnected)
             return
 
-        if not is_request and ui_client:
+        if not request and ui_client:
             #a bit of explanation:
             #normally these things are synchronized using xsettings, which we handle already
             #but non-posix clients have no such thing,
@@ -356,7 +370,7 @@ class ServerBase(ServerBaseClass):
                       self.session_name, self,
                       self.idle_add, self.timeout_add, self.source_remove,
                       self.setting_changed,
-                      self._socket_dir, self.unix_socket_paths, not is_request,
+                      self._socket_dir, self.unix_socket_paths, not request,
                       self.bandwidth_limit, self.bandwidth_detection,
                       )
         log("process_hello clientconnection=%s", ss)
@@ -369,7 +383,7 @@ class ServerBase(ServerBaseClass):
         self._server_sources[proto] = ss
         add_work_item(self.mdns_update)
         #process ui half in ui thread:
-        send_ui = ui_client and not is_request
+        send_ui = ui_client and not request
         self.idle_add(self._process_hello_ui, ss, c, auth_caps, send_ui, share_count)
 
     def get_client_connection_class(self, caps):
@@ -556,10 +570,10 @@ class ServerBase(ServerBaseClass):
 
     def send_hello_info(self, proto):
         self.wait_for_threaded_init()
-        start = monotonic_time()
+        start = monotonic()
         def cb(proto, info):
             self.do_send_info(proto, info)
-            end = monotonic_time()
+            end = monotonic()
             log.info("processed info request from %s in %ims",
                      proto._conn, (end-start)*1000)
         self.get_all_info(cb, proto, None)
@@ -579,24 +593,14 @@ class ServerBase(ServerBaseClass):
 
     def get_info(self, proto=None, client_uuids=None) -> dict:
         log("ServerBase.get_info%s", (proto, client_uuids))
-        start = monotonic_time()
-        info = ServerCore.get_info(self, proto)
+        start = monotonic()
         if client_uuids:
             sources = [ss for ss in self._server_sources.values() if ss.uuid in client_uuids]
         else:
             sources = tuple(self._server_sources.values())
         log("info-request: sources=%s", sources)
-        dgi = self.do_get_info(proto, sources)
-        #ugly alert: merge nested dictionaries,
-        #ie: do_get_info may return a dictionary for "server" and we already have one,
-        # so we update it with the new values
-        for k,v in dgi.items():
-            cval = info.get(k)
-            if cval is None:
-                info[k] = v
-                continue
-            cval.update(v)
-        log("ServerBase.get_info took %.1fms", 1000.0*(monotonic_time()-start))
+        info = self.do_get_info(proto, sources)
+        log("ServerBase.get_info took %.1fms", 1000.0*(monotonic()-start))
         return info
 
     def get_packet_handlers_info(self) -> dict:
@@ -617,18 +621,20 @@ class ServerBase(ServerBaseClass):
         return i
 
     def do_get_info(self, proto, server_sources=None) -> dict:
-        start = monotonic_time()
+        log("ServerBase.do_get_info%s", (proto, server_sources))
+        start = monotonic()
         info = {}
         def up(prefix, d):
             merge_dicts(info, {prefix : d})
 
         for c in SERVER_BASES:
             try:
+                cstart = monotonic()
                 merge_dicts(info, c.get_info(self, proto))
-            except Exception as e:
-                log("do_get_info%s", (proto, server_sources), exc_info=True)
-                log.error("Error collecting information from %s", c)
-                log.error(" %s", e)
+                cend = monotonic()
+                log("%s.get_info(%s) took %ims", c, proto, int(1000*(cend-cstart)))
+            except Exception:
+                log.error("Error collecting information from %s", c, exc_info=True)
 
         up("features",  self.get_features_info())
         up("network", {
@@ -655,7 +661,7 @@ class ServerBase(ServerBaseClass):
                 sinfo["ui-driver"] = self.ui_driver==ss.uuid
                 cinfo[i] = sinfo
             up("client", cinfo)
-        log("ServerBase.do_get_info took %ims", (monotonic_time()-start)*1000)
+        log("ServerBase.do_get_info took %ims", (monotonic()-start)*1000)
         return info
 
 
@@ -741,17 +747,6 @@ class ServerBase(ServerBaseClass):
         httplog("scripts=%s", scripts)
         return scripts
 
-    def cancel_http_stream_check_timers(self):
-        for ts in tuple(self.http_stream_check_timers.keys()):
-            v = self.http_stream_check_timers.pop(ts, None)
-            if v:
-                timer, stop = v
-                self.source_remove(timer)
-                try:
-                    stop()
-                except Exception:
-                    httplog("error on %s", stop, exc_info=True)
-
 
     ######################################################################
     # client connections:
@@ -776,6 +771,10 @@ class ServerBase(ServerBaseClass):
         return source
 
     def cleanup_source(self, source):
+        try:
+            ptype = source.protocol.TYPE
+        except (AttributeError, ValueError):
+            ptype = "xpra"
         self.server_event("connection-lost", source.uuid)
         remaining_sources = tuple(self._server_sources.values())
         if self.ui_driver==source.uuid:
@@ -785,7 +784,7 @@ class ServerBase(ServerBaseClass):
                 self.set_ui_driver(None)
         source.close()
         netlog("cleanup_source(%s) remaining sources: %s", source, remaining_sources)
-        netlog.info("xpra client %i disconnected.", source.counter)
+        netlog.info("%s client %i disconnected.", ptype, source.counter)
         has_client = len(remaining_sources)>0
         if not has_client:
             self.idle_add(self.last_client_exited)
