@@ -1,12 +1,12 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2012-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2012-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 
-from xpra.server.window.batch_config import DamageBatchConfig
+from xpra.server.window import batch_config
 from xpra.server.shadow.root_window_model import RootWindowModel
 from xpra.notifications.common import parse_image_path
 from xpra.platform.gui import get_native_notifier_classes, get_wm_name
@@ -37,7 +37,7 @@ if server_features.rfb:
 class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
 
     def __init__(self, root_window, capture=None):
-        SHADOWSERVER_BASE_CLASS.__init__(self)
+        super().__init__()
         self.capture = capture
         self.root = root_window
         self.mapped = []
@@ -50,8 +50,8 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
         self.pointer_last_position = None
         self.pointer_poll_timer = None
         self.last_cursor_data = None
-        DamageBatchConfig.ALWAYS = True             #always batch
-        DamageBatchConfig.MIN_DELAY = 50            #never lower than 50ms
+        batch_config.ALWAYS = True             #always batch
+        batch_config.MIN_DELAY = 50            #never lower than 50ms
 
     def init(self, opts):
         if SHADOWSERVER_BASE_CLASS!=object:
@@ -91,7 +91,7 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
         log("get_wm_name()=%s", self.session_name)
 
     def get_server_mode(self):
-        return "GTK3 shadow"
+        return "shadow"
 
     def print_screen_info(self):
         if not server_features.display:
@@ -132,8 +132,8 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
         #we export the whole desktop as a window:
         return 0, 0
 
-    def watch_keymap_changes(self):
-        pass
+    def _keys_changed(self, *_args):
+        log.info("keymap has been changed")
 
     def timeout_add(self, *args):
         #usually done via gobject
@@ -343,9 +343,8 @@ class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
                     from PIL import Image
                     img = Image.frombuffer("RGBA", (w, h), pixels, "raw", "BGRA", 0, 1)
                     img.save("cursor-%#x.png" % serial, format="PNG")
-            for ss in self._server_sources.values():
-                if hasattr(ss, "send_cursor"):
-                    ss.send_cursor()
+            for ss in self.window_sources():
+                ss.send_cursor()
 
     def do_get_cursor_data(self):
         #this method is overriden in subclasses with platform specific code
