@@ -1,21 +1,21 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 from math import sqrt
-from time import sleep
+from time import sleep, monotonic
 
 from xpra.server.source.stub_source_mixin import StubSourceMixin
-from xpra.server.window.batch_config import DamageBatchConfig
+from xpra.server.window import batch_config
 from xpra.server.server_core import ClientException
 from xpra.codecs.video_helper import getVideoHelper
 from xpra.codecs.codec_constants import video_spec
 from xpra.net.compression import use
-from xpra.os_util import monotonic_time, strtobytes
+from xpra.os_util import bytestostr
 from xpra.server.background_worker import add_work_item
 from xpra.util import csv, typedict, envint
 from xpra.log import Logger
@@ -25,28 +25,27 @@ proxylog = Logger("proxy")
 statslog = Logger("stats")
 
 MIN_PIXEL_RECALCULATE = envint("XPRA_MIN_PIXEL_RECALCULATE", 2000)
+DEFAULT_VREFRESH = envint("XPRA_DEFAULT_VREFRESH", 100)
 
 
-"""
-Store information about the client's support for encodings.
-Runs the encode thread.
-"""
 class EncodingsMixin(StubSourceMixin):
+    """
+    Store information about the client's support for encodings.
+    Runs the encode thread.
+    """
 
     @classmethod
     def is_needed(cls, caps : typedict) -> bool:
         return bool(caps.strtupleget("encodings"))
 
-
     def init_state(self):
         self.wants_encodings = False
         self.wants_features = False
 
         #contains default values, some of which may be supplied by the client:
-        self.default_batch_config = DamageBatchConfig()
+        self.default_batch_config = batch_config.DamageBatchConfig()
         self.global_batch_config = self.default_batch_config.clone()      #global batch config
 
-        self.vrefresh = -1
         self.supports_transparency = False
         self.encoding = None                        #the default encoding for all windows
         self.encodings = ()                         #all the encodings supported by the client
@@ -61,7 +60,6 @@ class EncodingsMixin(StubSourceMixin):
 
         self.zlib = True
         self.lz4 = use("lz4")
-        self.lzo = use("lzo")
 
         #for managing the recalculate_delays work:
         self.calculate_window_pixels = {}
@@ -73,6 +71,7 @@ class EncodingsMixin(StubSourceMixin):
         #new encoders, so we must make a deep copy to preserve the original
         #which may be used by other clients (other ServerSource instances)
         self.video_helper = getVideoHelper().clone()
+        self.cuda_device_context = None
 
 
     def init_from(self, _protocol, server):
@@ -85,8 +84,14 @@ class EncodingsMixin(StubSourceMixin):
         self.default_speed          = server.default_speed
         self.default_min_speed      = server.default_min_speed
 
+    def reinit_encodings(self, server):
+        self.server_core_encodings  = server.core_encodings
+        self.server_encodings       = server.encodings
+
     def cleanup(self):
         self.cancel_recalculate_timer()
+        if self.cuda_device_context:
+            self.queue_encode((False, self.free_cuda_device_context))
         #Warning: this mixin must come AFTER the window mixin!
         #to make sure that it is safe to add the end of queue marker:
         #(all window sources will have stopped queuing data)
@@ -94,6 +99,11 @@ class EncodingsMixin(StubSourceMixin):
         #this should be a noop since we inherit an initialized helper:
         self.video_helper.cleanup()
 
+    def free_cuda_device_context(self):
+        cdd = self.cuda_device_context
+        if cdd:
+            self.cuda_device_context = None
+            cdd.free()
 
     def all_window_sources(self):
         #we can't assume that the window mixin is loaded:
@@ -120,10 +130,10 @@ class EncodingsMixin(StubSourceMixin):
         self.calculate_timer = 0
         if self.is_closed():
             return
-        now = monotonic_time()
+        now = monotonic()
         self.calculate_last_time = now
         p = self.protocol
-        if not p:
+        if not p or p.is_closed():
             return
         conn = p._conn
         if not conn:
@@ -199,7 +209,7 @@ class EncodingsMixin(StubSourceMixin):
         if self.calculate_timer:
             #already due
             return
-        delta = monotonic_time() - self.calculate_last_time
+        delta = monotonic() - self.calculate_last_time
         RECALCULATE_DELAY = 1.0           #1s
         if delta>RECALCULATE_DELAY:
             add_work_item(self.recalculate_delays)
@@ -245,27 +255,21 @@ class EncodingsMixin(StubSourceMixin):
         #general features:
         self.zlib = c.boolget("zlib", True)
         self.lz4 = c.boolget("lz4", False) and use("lz4")
-        self.lzo = c.boolget("lzo", False) and use("lzo")
         self.brotli = c.boolget("brotli", False) and use("brotli")
-        log("compressors: zlib=%s, lz4=%s, lzo=%s, brotli=%s",
-            self.zlib, self.lz4, self.lzo, self.brotli)
+        log("compressors: zlib=%s, lz4=%s, brotli=%s",
+            self.zlib, self.lz4, self.brotli)
 
-        self.vrefresh = c.intget("vrefresh", -1)
-
-        #assume 50Hz:
-        ms_per_frame = 1000//100
-        if 30<=self.vrefresh<=500:
-            #looks like a valid vrefresh value, use it:
-            ms_per_frame = 1000//self.vrefresh
-        default_min_delay = max(DamageBatchConfig.MIN_DELAY, ms_per_frame)
+        delay = batch_config.START_DELAY
         dbc = self.default_batch_config
-        dbc.always      = bool(batch_value("always", DamageBatchConfig.ALWAYS))
-        dbc.min_delay   = batch_value("min_delay", default_min_delay, 0, 1000)
-        dbc.max_delay   = batch_value("max_delay", DamageBatchConfig.MAX_DELAY, 1, 15000)
-        dbc.max_events  = batch_value("max_events", DamageBatchConfig.MAX_EVENTS)
-        dbc.max_pixels  = batch_value("max_pixels", DamageBatchConfig.MAX_PIXELS)
-        dbc.time_unit   = batch_value("time_unit", DamageBatchConfig.TIME_UNIT, 1)
-        dbc.delay       = batch_value("delay", DamageBatchConfig.START_DELAY, 0)
+        dbc.always      = bool(batch_value("always", batch_config.ALWAYS))
+        dbc.min_delay   = batch_value("min_delay", batch_config.MIN_DELAY, 0, 1000)
+        dbc.max_delay   = batch_value("max_delay", batch_config.MAX_DELAY, 1, 15000)
+        dbc.max_events  = batch_value("max_events", batch_config.MAX_EVENTS)
+        dbc.max_pixels  = batch_value("max_pixels", batch_config.MAX_PIXELS)
+        dbc.time_unit   = batch_value("time_unit", batch_config.TIME_UNIT, 1)
+        self.vrefresh = c.intget("vrefresh", -1)
+        dbc.match_vrefresh(self.vrefresh)
+        dbc.delay       = batch_value("delay", delay, dbc.min_delay)
         log("default batch config: %s", dbc)
 
         #encodings:
@@ -307,18 +311,18 @@ class EncodingsMixin(StubSourceMixin):
         #2: standardized encoding options:
         for k in c.keys():
             #yaml gives us str..
-            k = strtobytes(k)
-            if k.startswith(b"theme.") or k.startswith(b"encoding.icons."):
-                self.icons_encoding_options[k.replace(b"encoding.icons.", b"").replace(b"theme.", b"")] = c.get(k)
-            elif k.startswith(b"encoding."):
-                stripped_k = k[len(b"encoding."):]
-                if stripped_k in (b"transparency",
-                                  b"rgb_zlib", b"rgb_lz4", b"rgb_lzo",
+            k = bytestostr(k)
+            if k.startswith("theme.") or k.startswith("encoding.icons."):
+                self.icons_encoding_options[k.replace("encoding.icons.", "").replace("theme.", "")] = c.get(k)
+            elif k.startswith("encoding."):
+                stripped_k = k[len("encoding."):]
+                if stripped_k in ("transparency",
+                                  "rgb_zlib", "rgb_lz4",
                                   ):
                     v = c.boolget(k)
-                elif stripped_k in (b"initial_quality", b"initial_speed",
-                                    b"min-quality", b"quality",
-                                    b"min-speed", b"speed"):
+                elif stripped_k in ("initial_quality", "initial_speed",
+                                    "min-quality", "quality",
+                                    "min-speed", "speed"):
                     v = c.intget(k)
                 else:
                     v = c.get(k)
@@ -356,6 +360,26 @@ class EncodingsMixin(StubSourceMixin):
         log("default encoding options: %s", self.default_encoding_options)
         self.auto_refresh_delay = c.intget("auto_refresh_delay", 0)
 
+        #are we going to need a cuda context?
+        if getattr(self, "mmap_enabled", False):
+            #not with mmap!
+            return
+        common_encodings = tuple(x for x in self.encodings if x in self.server_encodings)
+        from xpra.codecs.loader import has_codec
+        if (has_codec("nvenc") and "h264" in self.core_encodings or "h265" in self.core_encodings) \
+        or ("jpeg" in common_encodings and has_codec("enc_nvjpeg")):
+            cudalog = Logger("cuda")
+            try:
+                from xpra.codecs.cuda_common.cuda_context import get_device_context
+                self.cuda_device_context = get_device_context(self.encoding_options)
+                cudalog("cuda_device_context=%s", self.cuda_device_context)
+            except Exception as e:
+                cudalog("failed to get a cuda device context using encoding options %s",
+                    self.encoding_options, exc_info=True)
+                cudalog.error("Error: failed to allocate a CUDA context:")
+                cudalog.error(" %s", e)
+                cudalog.error(" NVJPEG and NVENC will not be available")
+
     def print_encoding_info(self):
         log("print_encoding_info() core-encodings=%s, server-core-encodings=%s",
             self.core_encodings, self.server_core_encodings)
@@ -374,7 +398,7 @@ class EncodingsMixin(StubSourceMixin):
 
     def parse_proxy_video(self):
         self.wait_for_threaded_init()
-        from xpra.codecs.enc_proxy.encoder import Encoder
+        from xpra.codecs.enc_proxy.encoder import Encoder  #pylint: disable=import-outside-toplevel
         proxy_video_encodings = self.encoding_options.get("proxy.video.encodings")
         proxylog("parse_proxy_video() proxy.video.encodings=%s", proxy_video_encodings)
         for encoding, colorspace_specs in proxy_video_encodings.items():
@@ -430,6 +454,7 @@ class EncodingsMixin(StubSourceMixin):
                 log.warn(" but it only supports: %s" % csv(self.encodings))
             if encoding not in self.server_encodings:
                 log.error("Error: encoding %s is not supported by this server", encoding)
+                log.error(" server encodings: %s", csv(self.server_encodings))
                 encoding = None
         if not encoding:
             encoding = "auto"
@@ -456,8 +481,6 @@ class EncodingsMixin(StubSourceMixin):
         info = {
                 "auto_refresh"      : self.auto_refresh_delay,
                 "lz4"               : self.lz4,
-                "lzo"               : self.lzo,
-                "vertical-refresh"  : self.vrefresh,
                 }
         ieo = dict(self.icons_encoding_options)
         ieo.pop("default.icons", None)
@@ -510,8 +533,8 @@ class EncodingsMixin(StubSourceMixin):
             log("batch config updated for window %s: %s", wid, ws.batch_config)
 
     def make_batch_config(self, wid : int, window):
-        batch_config = self.default_batch_config.clone()
-        batch_config.wid = wid
+        config = self.default_batch_config.clone()
+        config.wid = wid
         #scale initial delay based on window size
         #(the global value is normalized to 1MPixel)
         #but use sqrt to smooth things and prevent excesses
@@ -521,7 +544,7 @@ class EncodingsMixin(StubSourceMixin):
         w, h = window.get_dimensions()
         if dpm>=0:
             ratio = sqrt(1000000.0 / (w*h))
-            batch_config.delay = max(batch_config.min_delay, min(batch_config.max_delay, int(dpm * sqrt(ratio))))
+            config.delay = max(config.min_delay, min(config.max_delay, int(dpm * sqrt(ratio))))
         log("make_batch_config(%i, %s) global delay per megapixel=%i, new window delay for %ix%i=%s",
-                 wid, window, dpm, w, h, batch_config.delay)
-        return batch_config
+                 wid, window, dpm, w, h, config.delay)
+        return config
